#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy._deprecated._mbase import ModelBase


class SIR(ModelBase):
    """
    SIR model.

    Args:
        population (int): total population
        rho (float)
        sigma (float)
    """
    # Model name
    NAME = "SIR"
    # names of parameters
    PARAMETERS = ["rho", "sigma"]
    DAY_PARAMETERS = ["1/beta [day]", "1/gamma [day]"]
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = {
        "x": ModelBase.S,
        "y": ModelBase.CI,
        "z": ModelBase.FR
    }
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array([1, 1, 1])
    # Variables that increases monotonically
    VARS_INCREASE = [ModelBase.FR]
    # Example set of parameters and initial values
    EXAMPLE = {
        ModelBase.STEP_N: 180,
        ModelBase.N.lower(): 1_000_000,
        ModelBase.PARAM_DICT: {
            "rho": 0.2, "sigma": 0.075,
        },
        ModelBase.Y0_DICT: {
            ModelBase.S: 999_000, ModelBase.CI: 1000, ModelBase.FR: 0,
        },
    }

    @deprecate(old="SIR", new="SIRModel", version="2.24.0-xi")
    def __init__(self, population, rho, sigma):
        # Total population
        self.population = Validator(population, "population").int(value_range=(1, None))
        # Non-dim parameters
        self.rho = rho
        self.sigma = sigma
        self.non_param_dict = {"rho": rho, "sigma": sigma}

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.

        Args:
            t (int): time steps
            X (numpy.array): values of th model variables

        Returns:
            (np.array)
        """
        n = self.population
        s, i, *_ = X
        dsdt = 0 - self.rho * s * i / n
        drdt = self.sigma * i
        didt = 0 - dsdt - drdt
        return np.array([dsdt, didt, drdt])

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.

        Returns:
            float
        """
        try:
            rt = self.rho / self.sigma
        except ZeroDivisionError:
            return None
        return round(rt, 2)

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        try:
            return {
                "1/beta [day]": int(tau / 24 / 60 / self.rho),
                "1/gamma [day]": int(tau / 24 / 60 / self.sigma)
            }
        except (ZeroDivisionError, ValueError):
            return {p: None for p in self.DAY_PARAMETERS}

    @classmethod
    def convert(cls, data, tau):
        """
        Divide dates by tau value [min] and convert variables to model-specialized variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min] or None (skip division by tau values)

        Returns:
            pandas.DataFrame:
                Index
                    - Date (pd.Timestamp): Observation date (available when @tau is None)
                    - t (int): time steps (available when @tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
        """
        # Convert to tau-free if tau was specified
        df = cls._convert(data, tau)
        # Conversion of variables
        df[cls.FR] = df[cls.F] + df[cls.R]
        return df.loc[:, [cls.S, cls.CI, cls.FR]]

    @classmethod
    def convert_reverse(cls, converted_df, start, tau):
        """
        Calculate date with tau and start date, and restore Susceptible/Infected/"Fatal or Recovered".

        Args:
            converted_df (pandas.DataFrame):
                Index
                    t: Dates divided by tau value (time steps)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal or Recovered (int): the number of fatal/recovered cases
            start (pd.Timestamp): start date of simulation, like 14Apr2021
            tau (int): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        # Calculate date with tau and start date
        df = cls._convert_reverse(converted_df, start, tau)
        # Conversion of variables
        df[cls.F] = 0
        df[cls.R] = df.loc[:, cls.FR]
        return df.loc[:, [cls.DATE, cls.S, cls.CI, cls.F, cls.R]]

    @classmethod
    def guess(cls, data, tau, q=0.5):
        """
        With (X, dX/dt) for X=S, I, R, guess parameter values.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min]
            q (float or tuple(float,)): the quantile(s) to compute, value(s) between (0, 1)

        Returns:
            dict(str, float or pandas.Series): guessed parameter values with the quantile(s)

        Note:
            We can guess parameter values with difference equations as follows.
            - rho = - n * (dS/dt) / S / I
            - sigma = (dR/dt) / I
        """
        # Convert to tau-free and model-specialized dataset
        df = cls.convert(data=data, tau=tau)
        # Remove negative values and set variables
        df = df.loc[(df[cls.S] > 0) & (df[cls.CI] > 0)]
        n = df.loc[df.index[0], [cls.S, cls.CI, cls.FR]].sum()
        # Calculate parameter values with difference equation and tau-free data
        rho_series = 0 - n * df[cls.S].diff() / df[cls.S] / df[cls.CI]
        sigma_series = df[cls.FR].diff() / df[cls.CI]
        # Guess representative values
        return {
            "rho": cls._clip(rho_series.quantile(q=q), 0, 1),
            "sigma": cls._clip(sigma_series.quantile(q=q), 0, 1),
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from optuna.distributions import CategoricalDistribution, LogUniformDistribution
from optuna.integration import OptunaSearchCV
from sklearn.model_selection import TimeSeriesSplit
from sklearn.multioutput import MultiOutputRegressor
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from sklearn.svm import SVR
from covsirphy._deprecated.regbase import _RegressorBase
from covsirphy._deprecated.reg_rate_converter import _RateConverter
from covsirphy._deprecated.reg_feature_selector import _FeatureSelector


class _ParamSVRegressor(_RegressorBase):
    """
    Predict parameter values of ODE models with epsilon-support vector regressor.

    Args:
        - X_train (pandas.DataFrame): X for training with time index
        - X_test (pandas.DataFrame): X for test with time index
        - Y_train (pandas.DataFrame): Y for training with time index
        - Y_test (pandas.DataFrame): Y for test with time index
        - X_target (pandas.DataFrame): X for prediction with time index
    """
    # Description of regressor
    DESC = "Indicators -> Parameters with Epsilon-Support Vector Regressor"

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        """
        # Paramters of the steps
        param_grid = {
            "converter__to_convert": CategoricalDistribution([True, False]),
            "regressor__estimator__kernel": CategoricalDistribution(["linear", "rbf"]),
            "regressor__estimator__C": LogUniformDistribution(2 ** (-10), 2 ** 10),
            "regressor__estimator__gamma": LogUniformDistribution(2 ** (-10), 1),
            "regressor__estimator__epsilon": LogUniformDistribution(2 ** (-20), 1),
        }
        # Fit with pipeline
        steps = [
            ("converter", _RateConverter()),
            ("scaler", MinMaxScaler()),
            ("selector", _FeatureSelector(corr_threshold=0.8)),
            ("regressor", MultiOutputRegressor(SVR())),
        ]
        tscv = TimeSeriesSplit(n_splits=5).split(self._X_train)
        pipeline = OptunaSearchCV(
            Pipeline(steps=steps), param_grid, cv=tscv, random_state=0, n_trials=10)
        pipeline.fit(self._X_train, self._Y_train)
        # Update regressor
        self._pipeline = pipeline
        # Update param: pipeline.best_estimator_.named_steps.regressor.estimators_
        self._param.update(**{k: type(v) for (k, v) in steps})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta
import itertools
import warnings
import numpy as np
import pandas as pd
import ruptures as rpt
from covsirphy.util.error import SubsetNotFoundError, UnExpectedValueError, deprecate
from covsirphy.util.error import NotRegisteredMainError, NotRegisteredExtraError
from covsirphy.util.term import Term
from covsirphy.cleaning.cbase import CleaningBase
from covsirphy._deprecated.country_data import CountryData
from covsirphy.cleaning.japan_data import JapanData
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated.oxcgrt import OxCGRTData
from covsirphy._deprecated.pcr_data import PCRData
from covsirphy._deprecated.population import PopulationData
from covsirphy._deprecated.vaccine_data import VaccineData
from covsirphy._deprecated.mobility_data import MobilityData


class DataHandler(Term):
    """
    Data handler for analysis.

    Args:
        country (str): country name
        province (str or None): province name
        kwargs: arguments of DataHandler.register()
    """
    # Deprecated
    __NAME_COUNTRY = "CountryData"
    __NAME_JAPAN = "JapanData"
    # Extra datasets {str: class}
    __NAME_OXCGRT = "OxCGRTData"
    __NAME_PCR = "PCRData"
    __NAME_VACCINE = "VaccineData"
    __NAME_MOBILE = "MobilityData"
    EXTRA_DICT = {
        __NAME_COUNTRY: CountryData,
        __NAME_JAPAN: JapanData,
        __NAME_OXCGRT: OxCGRTData,
        __NAME_PCR: PCRData,
        __NAME_VACCINE: VaccineData,
        __NAME_MOBILE: MobilityData,
    }

    @deprecate(old="DataHandler", version="2.24.0-xi")
    def __init__(self, country, province=None, **kwargs):
        # Details of the area name
        self._area_dict = {"country": str(country), "province": str(province or self.NA)}
        # Main dataset before complement
        main_cols = [self.DATE, self.C, self.CI, self.F, self.R, self.S]
        self._main_raw = pd.DataFrame(columns=main_cols)
        # Main dataset After complement
        self._main_df = pd.DataFrame(columns=main_cols)
        # Extra dataset
        self._extra_df = pd.DataFrame(columns=[self.DATE])
        # Population
        self._population = None
        # Complement
        self._jhu_data = None
        self._complemented = None
        self._comp_dict = {}
        # Date
        self._first_date = None
        self._last_date = None
        self._today = None
        # Register datasets: date and main columns will be set internally if main data available
        self.register(**kwargs)

    @classmethod
    def _ensure_date_order(cls, previous_date, following_date, name="following_date"):
        """
        Ensure that the order of dates.

        Args:
            previous_date (str or pandas.Timestamp): previous date
            following_date (str or pandas.Timestamp): following date
            name (str): name of @following_date

        Raises:
            ValueError: @previous_date > @following_date
        """
        previous_date = cls._ensure_date(previous_date)
        following_date = cls._ensure_date(following_date)
        p_str = previous_date.strftime(cls.DATE_FORMAT)
        f_str = following_date.strftime(cls.DATE_FORMAT)
        if previous_date <= following_date:
            return None
        raise ValueError(f"@{name} must be the same as/over {p_str}, but {f_str} was applied.")

    @property
    def main_satisfied(self):
        """
        bool: all main datasets were registered or not
        """
        return not self._main_raw.empty

    @property
    def complemented(self):
        """
        bool or str: whether complemented or not and the details

        Raises:
            NotRegisteredMainError: no information because JHUData was not registered
        """
        if not self.main_satisfied:
            raise NotRegisteredMainError(".register(jhu_data)")
        return self._complemented

    @property
    @deprecate("DataHandler.population property", version="2.19.1-lambda")
    def population(self):
        """
        int: population value

        Raises:
            NotRegisteredMainError: no information because JHUData was not registered
        """
        if self._population is None:
            raise NotRegisteredMainError(".register(jhu_data)")
        return self._population

    @property
    def first_date(self):
        """
        str or None: the first date of the records
        """
        return self._first_date.strftime(self.DATE_FORMAT)

    @property
    def last_date(self):
        """
        str or None: the last date of the records
        """
        return self._last_date.strftime(self.DATE_FORMAT)

    @property
    def today(self):
        """
        str or None: reference date to determine whether a phase is a past phase or a future phase
        """
        return self._today.strftime(self.DATE_FORMAT)

    @staticmethod
    def _ensure_instance(target, class_obj, name="target"):
        """
        Ensure the target is an instance of the class object.

        Args:
            target (instance): target to ensure
            parent (class): class object
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        s = f"@{name} must be an instance of {class_obj}, but {type(target)} was applied."
        if not isinstance(target, class_obj):
            raise TypeError(s)
        return target

    def register(self, jhu_data=None, population_data=None, extras=None):
        """
        Register datasets.

        Args:
            jhu_data (covsirphy.JHUData or None): object of records
            population_data (covsirphy.PopulationData or None): PopulationData object (deprecated)
            extras (list[covsirphy.CleaningBase] or None): extra datasets

        Raises:
            TypeError: non-data cleaning instance was included
            UnExpectedValueError: instance of un-expected data cleaning class was included as an extra dataset
        """
        # Main: JHUData
        if jhu_data is not None:
            self._ensure_instance(jhu_data, JHUData, name="jhu_data")
            try:
                self._main_raw = jhu_data.subset(**self._area_dict, recovered_min=0)
            except SubsetNotFoundError as e:
                raise e from None
            self._jhu_data = jhu_data
            self.switch_complement(whether=True)
        # Main: PopulationData
        if population_data is not None:
            warnings.warn(
                ".register(population_data) was deprecated because population values are included in JHUData.",
                DeprecationWarning,
                stacklevel=2
            )
            self._ensure_instance(population_data, PopulationData, name="population_data")
            self._population = population_data.value(**self._area_dict)
        # Extra datasets
        if extras is not None:
            self._register_extras(extras)

    @staticmethod
    def _ensure_list(target, candidates=None, name="target"):
        """
        Ensure the target is a sub-list of the candidates.

        Args:
            target (list[object]): target to ensure
            candidates (list[object] or None): list of candidates, if we have
            name (str): argument name of the target

        Returns:
            object: as-is target
        """
        if not isinstance(target, (list, tuple)):
            raise TypeError(f"@{name} must be a list or tuple, but {type(target)} was applied.")
        if candidates is None:
            return target
        # Check the target is a sub-list of candidates
        try:
            strings = [str(candidate) for candidate in candidates]
        except TypeError:
            raise TypeError(f"@candidates must be a list, but {candidates} was applied.") from None
        ok_list = [element in candidates for element in target]
        if all(ok_list):
            return target
        candidate_str = ", ".join(strings)
        raise KeyError(f"@{name} must be a sub-list of [{candidate_str}], but {target} was applied.") from None

    def _register_extras(self, extras):
        """
        Verify the extra datasets.

        Args:
            extras (list[covsirphy.CleaningBase]): extra datasets

        Raises:
            TypeError: non-data cleaning instance was included as an extra dataset
            UnExpectedValueError: instance of un-expected data cleaning class was included as an extra dataset
        """
        self._ensure_list(extras, name="extras")
        # Verify the datasets
        for (i, extra_data) in enumerate(extras, start=1):
            statement = f"{self.num2str(i)} extra dataset"
            # Check the data is a data cleaning class
            self._ensure_instance(extra_data, CleaningBase, name=statement)
            # Check the data can be accepted as an extra dataset
            if isinstance(extra_data, (CountryData, JapanData)):
                warnings.warn(
                    ".register(extras=[CountryData, JapanData]) was deprecated because its role is played by the other classes.",
                    DeprecationWarning,
                    stacklevel=2
                )
            if isinstance(extra_data, tuple(self.EXTRA_DICT.values())):
                continue
            raise UnExpectedValueError(
                name=statement, value=type(extra_data), candidates=list(self.EXTRA_DICT.keys()))
        # Register the datasets
        extra_df = self._extra_df.set_index(self.DATE)
        for (extra_data, data_class) in itertools.product(extras, self.EXTRA_DICT.values()):
            if isinstance(extra_data, data_class):
                try:
                    subset_df = extra_data.subset(**self._area_dict)
                except TypeError:
                    subset_df = extra_data.subset(country=self._area_dict["country"])
                except SubsetNotFoundError:
                    continue
                extra_df = extra_df.combine_first(subset_df.set_index(self.DATE))
        self._extra_df = extra_df.reset_index()

    def recovery_period(self):
        """
        Return representative value of recovery period of all countries.

        Raises:
            NotRegisteredMainError: JHUData was not registered

        Returns:
            int or None: recovery period [days]
        """
        if self._jhu_data is None:
            raise NotRegisteredMainError(".register(jhu_data)")
        return self._jhu_data.recovery_period

    def records_main(self):
        """
        Return records of the main datasets as a dataframe from the first date to the last date.

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - Date (pd.Timestamp): Observation date
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases ( > 0)
                    - Susceptible (int): the number of susceptible cases
        """
        if self._main_df.empty:
            raise NotRegisteredMainError(".register(jhu_data)")
        df = self._main_df.copy()
        df = df.loc[(df[self.DATE] >= self._first_date) & (df[self.DATE] <= self._last_date)]
        return df.reset_index(drop=True)

    def switch_complement(self, whether=None, **kwargs):
        """
        Switch whether perform auto complement or not.

        Args:
            whether (bool): if True and necessary, the number of cases will be complemented
            kwargs: the other arguments of JHUData.subset_complement()
        """
        if not whether:
            df = self._main_raw.copy()
            self._main_df = df.loc[df[self.R] > 0].reset_index(drop=True)
            self._complemented = False
            return
        self._comp_dict.update(kwargs)
        if self._jhu_data is None:
            return
        self._main_df, self._complemented = self._jhu_data.records(**self._area_dict, **self._comp_dict)
        self.timepoints()

    def show_complement(self, **kwargs):
        """
        Show the details of complement that was (or will be) performed for the records.

        Args:
            kwargs: keyword arguments of JHUDataComplementHandler() i.e. control factors of complement

        Raises:
            NotRegisteredMainError: JHUData was not registered

        Returns:
            pandas.DataFrame: as the same as JHUData.show_complement()
        """
        if self._jhu_data is None:
            raise NotRegisteredMainError(".register(jhu_data)")
        comp_dict = self._comp_dict.copy()
        comp_dict.update(kwargs)
        return self._jhu_data.show_complement(
            start_date=self._first_date, end_date=self._last_date, **self._area_dict, **comp_dict)

    def timepoints(self, first_date=None, last_date=None, today=None):
        """
        Set the range of data and reference date to determine past/future of phases.

        Args:
            first_date (str or None): the first date of the records or None (min date of main dataset)
            last_date (str or None): the first date of the records or None (max date of main dataset)
            today (str or None): reference date to determine whether a phase is a past phase or a future phase

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data

        Note:
            When @today is None, the reference date will be the same as @last_date (or max date).
        """
        df = self._main_df.copy()
        first_date = self._ensure_date(
            first_date, name="first_date", default=self._first_date or df[self.DATE].min())
        last_date = self._ensure_date(
            last_date, name="last_date", default=self._last_date or df[self.DATE].max())
        today = self._ensure_date(today, name="today", default=min(self._today or last_date, last_date))
        # Check the order of dates
        self._ensure_date_order(df[self.DATE].min(), first_date, name="first_date")
        self._ensure_date_order(last_date, df[self.DATE].max(), name="the last date before changing")
        self._ensure_date_order(first_date, today, name="today")
        self._ensure_date_order(today, last_date, name="last_date")
        # Set timepoints
        self._first_date = first_date
        self._last_date = last_date
        self._today = today

    def records_extras(self):
        """
        Return records of the extra datasets as a dataframe.

        Raises:
            NotRegisteredMainError: JHUData was not registered
            NotRegisteredExtraError: no extra datasets were registered

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - Date(pd.Timestamp): Observation date
                    - columns defined in the extra datasets
        """
        if self._main_df.empty:
            raise NotRegisteredMainError(".register(jhu_data)")
        if self._extra_df.empty:
            raise NotRegisteredExtraError(".register(jhu_data, extras=[...]) with extra datasets")
        # Get all subset
        df = self._extra_df.copy()
        # Remove columns which is included in the main datasets
        unused_set = set(self._main_df.columns) - {self.DATE}
        df = df.loc[:, ~df.columns.isin(unused_set)]
        # Data cleaning
        df = df.set_index(self.DATE).resample("D").last()
        df = df.fillna(method="ffill").fillna(0)
        # Subsetting by dates
        df = df.loc[self._first_date: self._last_date]
        # Convert float values to integer if values will not be changed
        for col in df.columns:
            converted2int = df[col].astype(np.int64)
            if np.array_equal(converted2int, df[col]):
                df[col] = converted2int
        return df.reset_index()

    def _records(self, main=True, extras=True):
        """
        Return records of the datasets as a dataframe.

        Args:
            main (bool): whether include main datasets or not
            extras (bool): whether include extra datasets or not

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            NotRegisteredExtraError: @extras is True and no extra datasets were registered
            ValueError: both of @main and @extras were False

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - Date(pd.Timestamp): Observation date
                    - if @main is True,
                        - Confirmed(int): the number of confirmed cases
                        - Infected(int): the number of currently infected cases
                        - Fatal(int): the number of fatal cases
                        - Recovered (int): the number of recovered cases ( > 0)
                        - Susceptible(int): the number of susceptible cases
                    - if @extra is True,
                        - columns defined in the extra datasets
        """
        if main and extras:
            main_df = self.records_main()
            extra_df = self.records_extras()
            return main_df.merge(extra_df, on=self.DATE)
        if main:
            return self.records_main()
        if extras:
            return self.records_extras()
        raise ValueError("Either @main or @extras must be True.")

    def records(self, main=True, extras=True, past=True, future=True):
        """
        Return records of the datasets as a dataframe.

        Args:
            main (bool): whether include main datasets or not
            extras (bool): whether include extra datasets or not
            past (bool): whether include past records or not
            future (bool): whether include future records or not

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            NotRegisteredExtraError: @extras is True and no extra datasets were registered
            ValueError: both of @main and @extras were False, or both of @past and @future were False

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - Date(pd.Timestamp): Observation date
                    - if @main is True,
                        - Confirmed(int): the number of confirmed cases
                        - Infected(int): the number of currently infected cases
                        - Fatal(int): the number of fatal cases
                        - Recovered (int): the number of recovered cases ( > 0)
                        - Susceptible(int): the number of susceptible cases
                    - if @extra is True,
                        - columns defined in the extra datasets
        """
        if past and future:
            return self._records(main=main, extras=extras)
        if not past and not future:
            raise ValueError("Either @past or @future must be True.")
        df = self._records(main=main, extras=extras).set_index(self.DATE)
        if past:
            return df.loc[:self._today].reset_index()
        if future:
            return df.loc[self._today + timedelta(days=1):].reset_index()

    def records_all(self):
        """
        Return registered all records of the datasets as a dataframe.

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns:
                    - Date(pd.Timestamp): Observation date
                    - Confirmed(int): the number of confirmed cases
                    - Infected(int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases ( > 0)
                    - Susceptible(int): the number of susceptible cases
                    - columns defined in the extra datasets
        """
        try:
            return self.records(main=True, extras=True, past=True, future=True)
        except NotRegisteredExtraError:
            return self.records(main=True, extras=False, past=True, future=True)

    def estimate_delay(self, indicator, target, min_size=7, use_difference=False, delay_name="Period Length"):
        """
        Estimate the average day [days] between the indicator and the target.
        We assume that the indicator impact on the target value with delay.
        All results will be returned with a dataframe.

        Args:
            indicator (str): indicator name, a column of any registered datasets
            target (str): target name, a column of any registered datasets
            min_size (int): minimum size of the delay period
            use_difference (bool): if True, use first discrete difference of target
            delay_name (str): column name of delay in the output dataframe

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            UserWarning: failed in calculating and returned the default value (recovery period)

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - (int or float): column defined by @indicator
                    - (int or float): column defined by @target
                    - (int): column defined by @delay_name [days]

        Note:
            - We use change point analysis of ruptures package. Refer to the documentation.
              https://centre-borelli.github.io/ruptures-docs/
            - When failed in calculation, recovery period will be returned after raising UserWarning.
        """
        output_cols = [target, indicator, delay_name]
        # Create dataframe with indicator and target
        record_df = self.records_all()
        self._ensure_list(
            [indicator, target], candidates=record_df.columns.tolist(), name="indicator and target")
        if use_difference:
            record_df[target] = record_df[target].diff()
        pivot_df = record_df.pivot_table(values=indicator, index=target)
        run_df = pivot_df.copy()
        # Convert index (target) to serial numbers
        serial_df = pd.DataFrame(np.arange(1, run_df.index.max() + 1, 1))
        serial_df.index += 1
        run_df = run_df.join(serial_df, how="outer")
        series = run_df.reset_index(drop=True).iloc[:, 0].dropna()
        # Detection with Ruptures using indicator values
        warnings.simplefilter("ignore", category=RuntimeWarning)
        algorithm = rpt.Pelt(model="rbf", jump=1, min_size=min_size)
        try:
            results = algorithm.fit_predict(series.values, pen=0.5)
        except ValueError:
            default_delay = self.recovery_period()
            warnings.warn(
                f"Delay days could not be estimated and delay set to default: {default_delay} [days]",
                UserWarning, stacklevel=2)
            return pd.DataFrame(columns=output_cols)
        # Convert the output of Ruptures to indicator values
        reset_series = series.reset_index(drop=True)
        reset_series.index += 1
        results_df = reset_series[results].reset_index()
        results_df = results_df.interpolate(method="linear").dropna().astype(np.float64)
        # Convert the indicator values to dates
        df = pd.merge_asof(
            results_df.sort_values(indicator),
            pivot_df.astype(np.float64).reset_index().sort_values(indicator),
            on=indicator, direction="nearest")
        # Calculate number of days between the periods
        df[delay_name] = df["index"].sort_values(ignore_index=True).diff()
        return df.loc[:, output_cols]

    @classmethod
    def _ensure_date(cls, target, name="date", default=None):
        """
        Ensure the format of the string.

        Args:
            target (str or pandas.Timestamp): string to ensure
            name (str): argument name of the string
            default (pandas.Timestamp or None): default value to return

        Returns:
            pandas.Timestamp or None: as-is the target or default value
        """
        if target is None:
            return default
        if isinstance(target, pd.Timestamp):
            return target.replace(hour=0, minute=0, second=0, microsecond=0)
        try:
            return pd.to_datetime(target).replace(hour=0, minute=0, second=0, microsecond=0)
        except ValueError as e:
            raise ValueError(f"{name} was not recognized as a date, {target} was applied.") from e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import country_converter as coco
import numpy as np
import pandas as pd
import wbdata
from covsirphy.util.error import SubsetNotFoundError
from covsirphy.cleaning.cbase import CleaningBase


class PopulationPyramidData(CleaningBase):
    """
    Population pyramid dataset.
    World Bank Group (2020), World Bank Open Data, https://data.worldbank.org/

    Args:
        filename (str or None): CSV filename to save the dataset
        force (bool): if True, always download the dataset from the server
        verbose (int): level of verbosity

    Returns:
        If @filename is None, empty dataframe will be set as raw data.
        If @citation is None, citation will be empty string.
    """
    # Indicators of the raw dataset
    AGE_KEYS = [
        "0004", "0509", "1014", "1519", "2024", "2529", "3034", "3539",
        "4044", "4549", "5054", "5559", "6064", "6569", "7579", "80UP",
    ]
    INDICATOR_DICT = {
        f"SP.POP.{age}.{sex}": f"{age[:2]}-{age[2:]}-{sex}"
        for age in AGE_KEYS for sex in ["MA", "FE"]
    }
    ELDEST = 122
    # Columns
    SEX = "Sex"
    YEAR = "Year"
    AGE = "Age"
    PYRAMID_COLS = [CleaningBase.COUNTRY, YEAR, SEX, AGE, CleaningBase.N]
    PORTION = "Per_total"

    def __init__(self, filename, force=False, verbose=1):
        Path(filename).parent.mkdir(exist_ok=True, parents=True)
        if Path(filename).exists() and not force:
            self._raw = pd.read_csv(filename)
        else:
            self._raw = pd.DataFrame(columns=self.PYRAMID_COLS)
        self._cleaned_df = self._raw.copy()
        self._citation = "World Bank Group (2020), World Bank Open Data, https://data.worldbank.org/"
        self._filename = filename
        self.verbose = verbose

    def _retrieve_from_server(self, country):
        """
        Retrieve the dataset of the country from the server.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame: retrieved data
                Index
                    reset index
                Columns
                    - Country (object): country name
                    - Year (int): year
                    - Sex (object): Female/Male
                    - Age (object): age
                    - Population (object): population value
        """
        if self.verbose:
            print(
                f"Retrieving population pyramid dataset ({country}) from https://data.worldbank.org/")
        # Retrieve from World Bank Open Data
        iso3_code = coco.convert(country, to="ISO3", not_found=None)
        try:
            df = wbdata.get_dataframe(
                self.INDICATOR_DICT, country=iso3_code, convert_date=True)
        except RuntimeError:
            raise SubsetNotFoundError(country=country) from None
        # Preprocessing (-> Country, Population, Min, Max, Sex, Year)
        df = df.stack().reset_index()
        df.insert(0, self.COUNTRY, country)
        df.columns = [self.COUNTRY, "Date", "Attribute", self.N]
        df2 = df["Attribute"].str.split("-", expand=True)
        df2.columns = ["Min", "Max", self.SEX]
        df = pd.concat([df.drop("Attribute", axis=1), df2], axis=1)
        df["Max"] = df["Max"].replace("UP", self.ELDEST)
        for col in [self.N, "Min", "Max"]:
            df[col] = pd.to_numeric(df[col], downcast="integer")
        df[self.SEX].replace({"FE": "Female", "MA": "Male"}, inplace=True)
        df[self.YEAR] = df["Date"].dt.year
        df = df.drop("Date", axis=1)
        # Preprocessing (-> Country, Year, Sex, Age, Population)
        df[self.AGE] = df[["Min", "Max"]].apply(
            lambda x: range(x[0], x[1] + 1), axis=1)
        df[self.N] = df[["Min", "Max", self.N]].apply(
            lambda x: x[2] / (x[1] - x[0] + 1), axis=1)
        df = df.explode(self.AGE).reset_index(drop=True)
        df[self.N] = df[self.N].astype(np.int64)
        return df.loc[:, self.PYRAMID_COLS]

    def retrieve(self, country):
        """
        Retrieve the dataset of the country from the local file or the server.

        Args:
            country (str): country name

        Returns:
            pandas.DataFrame: retrieved data
                Index
                    reset index
                Columns
                    - Country (pandas.Category): country name
                    - Year (int): year
                    - Sex (str): Female/Male
                    - Age (int): age
                    - Population (int): population value
        """
        if not self._cleaned_df.empty and country in self._cleaned_df[self.COUNTRY].unique():
            df = self._cleaned_df.copy()
            df = df.loc[df[self.COUNTRY] == country, :].reset_index(drop=True)
        else:
            # Retrieve from World Bank Open Data
            try:
                df = self._retrieve_from_server(country)
            except SubsetNotFoundError:
                raise SubsetNotFoundError(country=country) from None
            # Add to raw dataset
            self._cleaned_df = pd.concat([self._cleaned_df, df], ignore_index=True, axis=0)
            self._cleaned_df.to_csv(self._filename, index=False)
        # Data types
        cat_cols, int_cols = [self.COUNTRY, self.SEX], [self.AGE, self.N]
        df[cat_cols] = df[cat_cols].astype("category")
        df[int_cols] = df[int_cols].astype(np.int64)
        return df

    def cleaned(self):
        """
        Return the cleaned dataset.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Country (pandas.Category): country name
                    - Year (int): year
                    - Sex (str): Female/Male
                    - Age (int): age
                    - Population (int): population value
        """
        return self._cleaned_df

    def layer(self, country=None):
        raise NotImplementedError

    def subset(self, country, year=None, sex=None):
        """
        Return the subset.

        Args:
            country (str): country name
            year (int or None): year or None (the last records)
            sex (str): Female/Male or None (total)

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Age (int): age
                    - Population (int): population value
                    - Per_total (float): portion of the total
        """
        # Select by country
        df = self.retrieve(country=country)
        # Select by year
        year = year or df[self.YEAR].max()
        df = df.loc[df[self.YEAR] == year, :]
        # Select by sex
        if sex is not None:
            df = df.loc[df[self.SEX] == sex, :]
        df = df.drop([self.COUNTRY, self.YEAR, self.SEX], axis=1)
        # Calculate portion of the total
        df = pd.DataFrame(df.groupby(self.AGE).sum())
        df[self.PORTION] = df[self.N] / df[self.N].sum()
        return df.reset_index()

    def records(self, country, year=None, sex=None):
        """
        Return the subset.

        Args:
            country (str): country name
            year (int or None): year or None (the last records)
            sex (str): Female/Male or None (total)

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Age (int): age
                    - Population (int): population value
                    - Per_total (float): portion of the total
        """
        return self.subset(country=country, year=year, sex=sex)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy._deprecated._mbase import ModelBase


class SEWIRF(ModelBase):
    """
    SEWIR-F model.
    """
    # Model name
    NAME = "SEWIR-F"
    # names of parameters
    PARAMETERS = ["theta", "kappa", "rho1", "rho2", "rho3", "sigma"]
    DAY_PARAMETERS = [
        "alpha1 [-]", "1/alpha2 [day]",
        "1/beta1 [day]", "1/beta2 [day]", "1/beta3 [day]",
        "1/gamma [day]"
    ]
    # Variable names in (non-dim, dimensional) ODEs
    VAR_DICT = {
        "x1": ModelBase.S,
        "x2": ModelBase.E,
        "x3": ModelBase.W,
        "y": ModelBase.CI,
        "z": ModelBase.R,
        "w": ModelBase.F,
    }
    VARIABLES = list(VAR_DICT.values())
    # Weights of variables in parameter estimation error function
    WEIGHTS = np.array([0, 10, 10, 2, 0, 0])
    # Variables that increases monotonically
    VARS_INCREASE = [ModelBase.R, ModelBase.F]
    # Example set of parameters and initial values
    EXAMPLE = {
        ModelBase.STEP_N: 180,
        ModelBase.N.lower(): 1_000_000,
        ModelBase.PARAM_DICT: {
            "theta": 0.002, "kappa": 0.005, "rho1": 0.2, "sigma": 0.075,
            "rho2": 0.167, "rho3": 0.167,
        },
        ModelBase.Y0_DICT: {
            ModelBase.S: 994_000, ModelBase.E: 3000, ModelBase.W: 2000,
            ModelBase.CI: 1000, ModelBase.R: 0, ModelBase.F: 0,
        },
    }

    @deprecate(old="SEWIRF", new="SEWIRFModel", version="2.24.0-xi")
    def __init__(self, population, theta, kappa, rho1, rho2, rho3, sigma):
        """

        Args:
        @population (int): total population
        parameter values of non-dimensional ODE model
            - @theta (float)
            - @kappa (float)
            - @rho1 (float)
            - @rho2 (float)
            - @rho3 (float)
            - @sigma (float)
        """
        # Total population
        self.population = Validator(population, "population").int(value_range=(1, None))
        # Non-dim parameters
        self.theta = theta
        self.kappa = kappa
        self.rho1 = rho1
        self.rho2 = rho2
        self.rho3 = rho3
        self.sigma = sigma
        self.non_param_dict = {
            "theta": theta, "kappa": kappa,
            "rho1": rho1, "rho2": rho2, "rho3": rho3,
            "sigma": sigma
        }

    def __call__(self, t, X):
        """
        Return the list of dS/dt (tau-free) etc.

        Args:
            t (int): time steps
            X (numpy.array): values of th model variables

        Returns:
            (np.array)
        """
        n = self.population
        s, i, *_, e, w = X
        beta_swi = self.rho1 * s * (w + i) / n
        dsdt = 0 - beta_swi
        dedt = beta_swi - self.rho2 * e
        dwdt = self.rho2 * e - self.rho3 * w
        drdt = self.sigma * i
        dfdt = self.kappa * i + self.theta * self.rho3 * w
        didt = 0 - dsdt - drdt - dfdt - dedt - dwdt
        return np.array([dsdt, didt, drdt, dfdt, dedt, dwdt])

    def calc_r0(self):
        """
        Calculate (basic) reproduction number.

        Returns:
            float
        """
        try:
            rho = self.rho1 / self.rho2 * self.rho3
            rt = rho * (1 - self.theta) / (self.sigma + self.kappa)
        except ZeroDivisionError:
            return None
        return round(rt, 2)

    def calc_days_dict(self, tau):
        """
        Calculate 1/beta [day] etc.

        Args:
            param tau (int): tau value [min]

        Returns:
            dict[str, int]
        """
        try:
            return {
                "alpha1 [-]": round(self.theta, 3),
                "1/alpha2 [day]": int(tau / 24 / 60 / self.kappa),
                "1/beta1 [day]": int(tau / 24 / 60 / self.rho1),
                "1/beta2 [day]": int(tau / 24 / 60 / self.rho2),
                "1/beta3 [day]": int(tau / 24 / 60 / self.rho3),
                "1/gamma [day]": int(tau / 24 / 60 / self.sigma)
            }
        except (ZeroDivisionError, ValueError):
            return {p: None for p in self.DAY_PARAMETERS}

    @classmethod
    def convert(cls, data, tau):
        """
        Divide dates by tau value [min] and convert variables to model-specialized variables.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min] or None (skip division by tau values)

        Returns:
            pandas.DataFrame:
                Index
                    - Date (pd.Timestamp): Observation date (available when @tau is None)
                    - t (int): time steps (available when @tau is not None)
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Exposed (int): 0
                    - Waiting (int): 0
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
        """
        # Convert to tau-free if tau was specified
        df = cls._convert(data, tau)
        # Conversion of variables
        df[cls.E] = 0
        df[cls.W] = 0
        return df.loc[:, [cls.S, cls.E, cls.W, cls.CI, cls.R, cls.F]]

    @classmethod
    def convert_reverse(cls, converted_df, start, tau):
        """
        Calculate date with tau and start date, and restore Susceptible/Infected/Fatal/Recovered.

        Args:
            converted_df (pandas.DataFrame):
                Index
                    time steps: Dates divided by tau value
                Columns
                    - Susceptible (int): the number of susceptible cases
                    - Exposed (int): exposed and in latent period (without infectivity)
                    - Waiting (int): waiting for confirmation diagnosis (with infectivity)
                    - Infected (int): the number of currently infected cases
                    - Recovered (int): the number of recovered cases
                    - Fatal (int): the number of fatal cases
            start (pd.Timestamp): start date of simulation, like 14Apr2021
            tau (int): tau value [min]

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
        """
        # Calculate date with tau and start date
        df = cls._convert_reverse(converted_df, start, tau)
        # Conversion of variables
        df[cls.S] = df[cls.S] + df[cls.E] + df[cls.W]
        return df.loc[:, [cls.DATE, cls.S, cls.CI, cls.F, cls.R]]

    @classmethod
    def guess(cls, data, tau, q=0.5):
        """
        With (X, dX/dt) for X=S, I, R and so on, guess parameter values.
        This is not implemented for SEWIR-F model.

        Args:
            data (pandas.DataFrame):
                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Susceptible(int): the number of susceptible cases
                    - Infected (int): the number of currently infected cases
                    - Fatal(int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
            tau (int): tau value [min]
            q (float or tuple(float,)): the quantile(s) to compute, value(s) between (0, 1)
        """
        raise NotImplementedError(
            "SEWIR-F cannot be used for parameter estimation because we do not have records "
            "of Exposed and Waiting. Please use SIR-F model with `covsirphy.SIRF` class."
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
import pandas as pd
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term


class _DataCleaner(Term):
    """Class for data cleaning.

    Args:
        data (pandas.DataFrame): raw data
            Index
                reset index
            Column
                columns defined by @layers
                column defined by @date
                the other columns
        layers (list[str]): location layers of the data
        date (str): column name of observation dates of the data
    """

    def __init__(self, data, layers, date):
        self._df = data.copy()
        self._layers = Validator(layers, "layers").sequence()
        self._date = str(date)
        self._id_cols = [*self._layers, self._date]

    def all(self):
        """Return all available data.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Column
                    columns defined by @layers of _DataCleaner()
                    (pandas.Timestamp): observation dates defined by @date of _DataCleaner()
                    the other columns
        """
        return self._df

    def convert_date(self, **kwargs):
        """Convert dtype of date column to pandas.Timestamp.

        Args:
            **kwargs: keyword arguments of pandas.to_datetime() including "dayfirst (bool): whether date format is DD/MM or not"
        """
        df = self._df.copy()
        df[self._date] = pd.to_datetime(df[self._date], **kwargs).dt.round("D")
        with contextlib.suppress(TypeError):
            df[self._date] = df[self._date].dt.tz_convert(None)
        self._df = df.copy()

    def resample(self):
        """Resample records with dates.
        """
        self.convert_date()
        df = self._df.copy()
        grouped = df.set_index(self._date).groupby(self._layers, as_index=False)
        df = grouped.resample("D").ffill()
        self._df = df.reset_index().drop("level_0", errors="ignore", axis=1)

    def fillna(self):
        """Fill NA values with '-' (layers) and the previous values and 0.
        """
        df = self._df.copy()
        df.loc[:, self._layers] = df.loc[:, self._layers].astype(str).fillna(self.NA)
        for col in set(df.columns) - set(self._id_cols):
            df[col] = df.groupby(self._layers)[col].ffill().fillna(0)
        self._df = df.copy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from unidecode import unidecode
from covsirphy.util.term import Term
from covsirphy._deprecated.db_base import _RemoteDatabase


class _GoogleOpenData(_RemoteDatabase):
    """
    Access "COVID-19 Open Data by Google Cloud Platform".
    https://github.com/GoogleCloudPlatform/covid-19-open-data
    https://goo.gle/covid-19-open-data

    Args:
        filename (str): CSV filename to save records
        iso3 (str or None): ISO3 code of the country which must be included in the dataset or None (all available countries)
    """
    # URL for mobility data
    URL_M = "https://storage.googleapis.com/covid19-open-data/v3/mobility.csv"
    # URL for index data (country names etc.)
    URL_I = "https://storage.googleapis.com/covid19-open-data/v3/index.csv"
    # Citation
    CITATION = "O. Wahltinez and others (2020)," \
        " COVID-19 Open-Data: curating a fine-grained, global-scale data repository for SARS-CoV-2, " \
        " Work in progress, https://goo.gle/covid-19-open-data"
    # Column names and data types
    # {"name in database": "name defined in Term class"}
    _MOBILITY_COLS_RAW_INT = [
        "mobility_grocery_and_pharmacy",
        "mobility_parks",
        "mobility_transit_stations",
        "mobility_retail_and_recreation",
        "mobility_residential",
        "mobility_workplaces",
    ]
    MOBILITY_VARS = [v.capitalize() for v in _MOBILITY_COLS_RAW_INT]
    COL_DICT = {
        "date": Term.DATE,
        "iso_3166_1_alpha_3": Term.ISO3,
        Term.PROVINCE: Term.PROVINCE,
        **{v: v.capitalize() for v in _MOBILITY_COLS_RAW_INT},
    }

    def download(self, verbose):
        """
        Download the dataset from the server and set the list of primary sources.

        Args:
            verbose (int): level of verbosity

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    defined by the first values of self.COL_DICT.values()

        Note:
            If @verbose is equal to or over 1, how to show the list will be explained.
        """

        # Download datasets
        if verbose:
            print("Retrieving datasets from COVID-19 Open Data by Google Cloud Platform https://github.com/GoogleCloudPlatform/covid-19-open-data")
        # Index
        i_cols = ["location_key", "iso_3166_1_alpha_3", "subregion1_name", "subregion2_name"]
        i_df = pd.read_csv(self.URL_I, usecols=i_cols)
        # Mobility
        m_df = pd.read_csv(self.URL_M)
        m_df = (m_df.set_index(["date", "location_key"]) + 100).reset_index()
        # Combine data
        df = m_df.merge(i_df, how="left", on="location_key")
        # Location (iso3/province): remove city-level data
        df = df.loc[df["subregion2_name"].isna()]
        df[self.PROVINCE] = df["subregion1_name"].fillna(self.NA).apply(unidecode)
        return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd
from sklearn.linear_model import MultiTaskElasticNetCV
from sklearn.model_selection import TimeSeriesSplit
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import MinMaxScaler
from covsirphy._deprecated.regbase import _RegressorBase
from covsirphy._deprecated.reg_rate_converter import _RateConverter
from covsirphy._deprecated.reg_feature_selector import _FeatureSelector


class _ParamElasticNetRegressor(_RegressorBase):
    """
    Predict parameter values of ODE models with Elastic Net regression.

    Args:
        - X_train (pandas.DataFrame): X for training with time index
        - X_test (pandas.DataFrame): X for test with time index
        - Y_train (pandas.DataFrame): Y for training with time index
        - Y_test (pandas.DataFrame): Y for test with time index
        - X_target (pandas.DataFrame): X for prediction with time index
    """
    # Description of regressor
    DESC = "Indicators -> Parameters with Elastic Net"

    def _fit(self):
        """
        Fit regression model with training dataset, update self._pipeline and self._param.
        """
        # Model for Elastic Net regression
        tscv = TimeSeriesSplit(n_splits=5).split(self._X_train)
        cv = MultiTaskElasticNetCV(
            alphas=[0, 0.001, 0.01, 0.1, 1, 10, 100, 1000],
            l1_ratio=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
            cv=tscv, n_jobs=-1
        )
        # Fit with pipeline
        steps = [
            ("converter", _RateConverter(to_convert=False)),
            ("scaler", MinMaxScaler()),
            ("selector", _FeatureSelector(corr_threshold=0.8)),
            ("regressor", cv),
        ]
        pipeline = Pipeline(steps=steps)
        pipeline.fit(self._X_train, self._Y_train)
        reg_output = pipeline.named_steps.regressor
        # Update regressor
        self._pipeline = pipeline
        # Intercept/coef
        intercept_df = pd.DataFrame(
            reg_output.coef_, index=self._Y_train.columns, columns=self._X_train.columns)
        intercept_df.insert(0, "Intercept", None)
        intercept_df["Intercept"] = reg_output.intercept_
        # Update param
        param_dict = {
            **{k: type(v) for (k, v) in steps},
            "alpha": reg_output.alpha_,
            "l1_ratio": reg_output.l1_ratio_,
            "intercept": intercept_df,
            "coef": intercept_df,
        }
        self._param.update(param_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from itertools import chain
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.scenario import Scenario
from covsirphy._deprecated.example_data import ExampleData


class ModelValidator(Term):
    """
    Deprecated. Evaluate ODE models performance as follows.
    1. Select model parameter sets randomly
    2. Set user-defined/random phase duration
    3. Perform simulation with a specified ODE model
    4. Perform parameter estimation
    5. Compare the estimated parameters and the parameters produced with th 1st step
    6. Repeat trials (1 trial = from the 1st step to the 5th step)
    Small difference is expected in the 6th step.

    Args:
        tau (int): tau value [min]
        n_trials (int): the number of trials
        step_n (int or None): the number of steps in simulation (over 2) or None (randomly selected)
        seed (int): random seed

    Note:
        Population value and initial values are defined by model.EXAMPLE.
        Estimators know tau values before parameter estimation.
    """

    @deprecate(old="ModelValidator()", version="2.24.0-kappa")
    def __init__(self, tau=1440, n_trials=8, step_n=None, seed=0):
        self._tau = Validator(tau, "tau").tau(default=None)
        self._n_trials = Validator(n_trials, "n_trials").int(value_range=(1, None))
        self._seed = Validator(seed, "seed").int(value_range=(0, None))
        # list[int]: the number of steps for each trials
        if step_n is None:
            np.random.seed(seed)
            self._step_n_list = np.random.randint(5, 60, n_trials)
        else:
            Validator(step_n, "step_n").int(value_range=(3, None))
            self._step_n_list = [step_n for _ in range(n_trials)]
        # Validated models
        self.model_names = []
        # Dataframes of results created by ._get_result()
        self._results = []

    def run(self, model, timeout=180, allowance=(0.98, 1.02), n_jobs=-1):
        """
        Execute model validation.

        Args:
            model (covsirphy.ModelBase): ODE model
            timeout (int): time-out of run
            allowance (tuple(float, float)): the allowance of the predicted value
            n_jobs (int): the number of parallel jobs or -1 (CPU count)

        Returns:
            covsirphy.ModelValidator: self
        """
        model = Validator(model, "model").subclass(ModelBase)
        if model.NAME in self.model_names:
            raise ValueError(f"{model.NAME} has been validated.")
        self.model_names.append(model.NAME)
        # Setup: create parameter set, phase length and processor
        df = self._setup(model)
        # Parameter estimation
        scenarios = self._processor(model, df, timeout=timeout, allowance=allowance)
        # Get estimated parameters
        self._results.append(self._get_result(model, df, scenarios))
        return self

    def _setup(self, model):
        """
        Create a model parameter set randomly.

        Args:
            model (covsirphy.ModelBase): ODE model

        Returns:
            pandas.DataFrame:
                Index reset index
                Columns
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation
        """
        np.random.seed(self._seed)
        population = model.EXAMPLE[self.N.lower()]
        # Parameter set
        parameters = model.PARAMETERS[:]
        df = pd.DataFrame(np.random.rand(self._n_trials, len(parameters)), columns=parameters)
        # Reproduction number
        df[self.RT] = df.apply(lambda x: model(population, **x.to_dict()).calc_r0(), axis=1)
        # Tau value
        df[self.TAU] = self._tau
        # Step number
        df[self.STEP_N] = self._step_n_list
        # Return the setting
        return df

    def _processor(self, model, setting_df, timeout, allowance):
        """
        Generate multi-processor for parameter estimation,
        registering theoretical data and phase units.

        Args:
            model (covsirphy.ModelBase): ODE model
            setting_df (pandas.DataFrame):
                Index reset index
                Columns
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation
            timeout (int): time-out of run
            allowance (tuple(float, float)): the allowance of the predicted value

        Returns:
            list[covsirphy.Scenario]: list of Scenario instances
        """
        scenarios = []
        # Instance to save theoretical data
        example_data = ExampleData(tau=self._tau, start_date="01Jan2020")
        # Population values
        population = model.EXAMPLE[self.N.lower()]
        # Register data for each setting
        for (i, setting_dict) in enumerate(setting_df.to_dict(orient="records")):
            name = f"{model.NAME}_{i}"
            step_n = setting_dict[self.STEP_N]
            param_dict = {k: v for (k, v) in setting_dict.items() if k in model.PARAMETERS}
            # Add theoretical data
            example_data.add(
                model, step_n=step_n, country=name, param_dict=param_dict, population=population)
            # Phase unit
            snl = Scenario(country=name, auto_complement=False)
            snl.register(example_data)
            snl.add()
            snl.estimate(model, timeout=timeout, allowance=allowance)
            scenarios.append(snl)
        return scenarios

    def _get_result(self, model, setting_df, scenarios):
        """
        Show the result as a dataframe.

        Args:
            model (covsirphy.ModelBase): ODE model
            setting_df (pandas.DataFrame):
                Index reset index
                Columns
                    - (float): parameter values from 0 to 1.0
                    - Rt (float): reproduction number
                    - step_n (int): step number of simulation
            list[covsirphy.Scenario]: list of Scenario instances

        Returns:
            pandas.DataFrame:
                Index reset index
                Columns
                    - ID (str): ID, like SIR_0
                    - ODE (str): model name
                    - Rt (float): reproduction number set by ._setup() method
                    - Rt_est (float): estimated reproduction number
                    - rho etc. (float): parameter values set by ._setup() method
                    - rho_est etc. (float): estimated parameter values
                    - step_n (int): step number of simulation
                    - RMSLE (float): RMSLE score of parameter estimation
                    - Trials (int): the number of trials in parameter estimation
                    - Runtime (str): runtime of parameter estimation
        """
        df = setting_df.copy()
        results_df = pd.concat([snl.summary() for snl in scenarios], axis=0)
        df = df.join(results_df.reset_index(drop=True), how="left", rsuffix="_est")
        df = df.drop(model.DAY_PARAMETERS, axis=1, errors="ignore")
        cols_to_alternate = [self.RT, *model.PARAMETERS]
        cols_alternated = chain.from_iterable(
            zip(cols_to_alternate, [f"{col}_est" for col in cols_to_alternate]))
        df["ID"] = df[self.ODE].str.cat(df.index.astype("str"), sep="_")
        columns = ["ID", self.ODE, *cols_alternated, self.STEP_N, "RMSLE", self.TRIALS, self.RUNTIME]
        return df.loc[:, columns]

    def summary(self):
        """
        Show the summary of validation.

        Returns:
            pandas.DataFrame:
                Index
                    reset index
                Columns
                    - ID (str): ID, like SIR_0
                    - ODE (str): model name
                    - Rt (float): reproduction number set by ._setup() method
                    - Rt_est (float): estimated reproduction number
                    - rho etc. (float): parameter values set by ._setup() method
                    - rho_est etc. (float): estimated parameter values
                    - step_n (int): step number of simulation
                    - RMSLE (float): RMSLE score of parameter estimation
                    - Trials (int): the number of trials in parameter estimation
                    - Runtime (str): runtime of parameter estimation
        """
        df = pd.concat(self._results, ignore_index=True, sort=True)
        pre_cols = ["ID", self.ODE, self.RT, f"{self.RT}_est"]
        post_cols = [self.STEP_N, "RMSLE", self.TRIALS, self.RUNTIME]
        centers = list(set(df.columns) - set(pre_cols) - set(post_cols))
        centers_sorted = sorted(centers, key=lambda x: df.columns.tolist().index(x))
        return df[[*pre_cols, *centers_sorted, *post_cols]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
import warnings
import geopandas as gpd
from unidecode import unidecode
from covsirphy.util.validator import Validator
from covsirphy.util.filer import Filer
from covsirphy.util.term import Term


class _Geometry(Term):
    """Class to add geometry information to geo-spatial data.

    Args:
        data (pandas.DataFrame): geo-spatial data
            Index
                reset index
            Columns
                - column defined by @layer
                - the other columns
        layer (str): layer name which has location information
        directory (str): directory to save GeoJSON files
        verbose (int): level of verbosity of stdout
    """

    def __init__(self, data, layer, directory, verbose):
        self._df = Validator(data, f"{layer}-level data").dataframe(columns=[layer])
        self._layer = layer
        self._filer = Filer(directory=directory)
        self._verbose = Validator(verbose, "verbose").int(value_range=(0, None))

    def to_geopandas(self, iso3, natural_earth):
        """Add geometry information with GeoJSON file of "Natural Earth" GitHub repository to data.

        Args:
            iso3 (str or None): ISO3 code (for the province/city-level data) or None (country-level data)
            natural_earth (str or None): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository or None (automatically determined)

        Returns:
            geopandas.GeoDataFrame:
                Index:
                    - reset index
                Columns:
                    - columns included in @data of _Geometry()
                    - geometry: geometric information

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
            https://www.naturalearthdata.com/
            https://github.com/nvkelso/natural-earth-vector
            Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)
        """
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        gdf = self._natural_earth(title=natural_earth or self._natural_earth_parse_title(iso3=iso3))
        right_on = self.ISO3 if iso3 is None else "NAME"
        gdf = self._df.merge(gdf, how="left", left_on=self._layer, right_on=right_on)
        return gdf.rename(columns={f"{self.ISO3}_x": self.ISO3}).loc[:, [*self._df.columns.tolist(), "geometry"]]

    @staticmethod
    def _natural_earth_parse_title(iso3):
        """Find the best file title ("Natural Earth" GitHub repository) for the layer and country.

        Args:
            iso3 (str or None): ISO3 code (for the province/city-level data) or None (country-level data)

        Returns:
            str: parsed title
        """
        if iso3 is None:
            return "ne_110m_admin_0_countries"
        scale = "50m" if iso3 == "USA" else "10m"
        return f"ne_{scale}_admin_1_states_provinces"

    def _natural_earth(self, title):
        """Download GeoJSON files from "Natural Earth" GitHub repository.
        https://www.naturalearthdata.com/
        https://github.com/nvkelso/natural-earth-vector
        Natural Earth (Free vector and raster map data at naturalearthdata.com, Public Domain)

        Args:
            title (str): title of GeoJSON file (without extension) of "Natural Earth" GitHub repository

        Returns:
            geopandas.GeoDataFrame:
                Index:
                    - reset index
                Columns:
                    - ISO3: ISO3 codes
                    - NAME: Country names or province names
                    - geometry: geometric information

        Note:
            GeoJSON files are listed in https://github.com/nvkelso/natural-earth-vector/tree/master/geojson
        """
        file_dict = self._filer.geojson(title, driver="GeoJSON")
        filename = file_dict["filename"]
        if Path(filename).exists():
            gdf = gpd.read_file(filename)
        else:
            if self._verbose:
                print("Retrieving GIS data from Natural Earth https://www.naturalearthdata.com/")
            url = f"https://raw.githubusercontent.com/nvkelso/natural-earth-vector/master/geojson/{title}.geojson"
            gdf = gpd.read_file(url)
            gdf.rename(columns={"ISO_A3": self.ISO3, "adm0_a3": self.ISO3, "name": "NAME"}, inplace=True)
            gdf["NAME"] = gdf["NAME"].fillna(self.NA).apply(unidecode)
            gdf["NAME"] = gdf["NAME"].replace(
                {"Xizang": "Tibet", "Inner Mongol": "Inner Mongolia", "S. Sudan": "South Sudan",
                 "Fr. S. Antarctic Lands": "French Southern and Antarctic Lands"})
            gdf.loc[gdf[self.ISO3] == "HKG", "NAME"] = "Hong Kong"
            gdf.loc[gdf[self.ISO3] == "-99", self.ISO3] = self._to_iso3(gdf.loc[gdf[self.ISO3] == "-99", "NAME"])
            gdf[self.ISO3] = gdf[self.ISO3].replace({"MAC": "CHN"})
            gdf.to_file(**file_dict)
        return gdf.reindex(columns=[self.ISO3, "NAME", "geometry"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pca import pca
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.science._autots import _AutoTSHandler


class MLEngineer(Term):
    """
    Class for machine learning and preprocessing.

    Args:
        seed (int or None): random seed
        verbose (int): verbosity, 0: None, 1: Error, 2: Warning, 3: Info, 4: Debug, 5: Trace for PCA.
    """

    def __init__(self, seed=0, verbose=3):
        self._seed = Validator(seed, name="seed").int()
        self._verbose = Validator(verbose, name="verbose").int()

    def pca(self, X, n_components=0.95):
        """Perform PCA (principal component analysis) after standardization (Z-score normalization) with pca package.

        Args:
            X (pandas.DataFrame or None):
                Index
                    pandas.Timestamp: Observation date
                Columns
                    (int or float): observed values of the training vectors
            n_components (float or int): _the number of principal components or percentage of variance to cover at least

        Returns:
            dict of {str: object}: as the same as pca.pca().fit_transform()
                {"loadings": pandas.DataFrame}: structured dataframe containing loadings for PCs
                {"PC": pandas.DataFrame}: reduced dimensionality space, the Principal Components (PCs)
                    Index
                        pandas.Timestamp
                    COlumns
                        PC1, PC2,...
                {"explained_var": array-like}: explained variance for each fo the PCs (same ordering as the PCs)
                {"variance_ratio": array-like};: variance ratio
                {"model": object}: fitted model to be used for further usage of the model
                {"scaler": object}: scaler model
                {"pcp": int}: pcp
                {"topfeat": pandas.DataFrame}: top features
                    Index
                        reset index
                    Columns
                        PC (str): PC1, PC2,...
                        feature (str): feature name of X
                        loading (float): loading values
                        type (str): "best" or "weak
                {"outliers": pandas.DataFrame}: outliers
                    Index
                        pandas.Timestamp
                    Columns
                        y_proba (float)
                        y_score (float)
                        y_bool (bool)
                        y_bool_spe (bool)
                        y_score_spe (float)
                {"outlier_params": object}: parameter values of the model of finding outliers

        Note:
            Regarding pca package, please refer to https://github.com/erdogant/pca
        """
        Validator(X, name="X", accept_none=False).dataframe(time_index=True, empty_ok=False)
        model = pca(n_components=n_components, normalize=True, random_state=self._seed, verbose=self._verbose)
        return {**model.fit_transform(X), "model": model}

    def forecast(self, Y, days, X=None, **kwargs):
        """Forecast Y for given days with/without indicators (X).

        Args:
            Y (pandas.DataFrame):
                Index
                    pandas.Timestamp: Observation date
                Columns
                    observed and the target variables (int or float)
            X (pandas.DataFrame or None): indicators for regression or None (no indicators)
                Index
                    pandas.Timestamp: Observation date
                Columns
                    observed and the target variables (int or float)
            days (int): days to predict
            **kwargs: keyword arguments of autots.AutoTS() except for verbose, forecast_length (always the same as @days)

        Return:
           pandas.DataFrame:
                Index
                    pandas.Timestamp: Observation date, from the next date of Y.index to the ast predicted date
                Columns
                    observed and the target variables (int or float)

        Note:
            AutoTS package is developed at https://github.com/winedarksea/AutoTS
        """
        model = _AutoTSHandler(Y=Y, days=days, seed=self._seed, verbose=self._verbose, **kwargs)
        return model.fit(X=X).predict()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.util.error import deprecate
from covsirphy.util.term import Term


class Estimator(Term):
    """
    Deprecated. Hyperparameter optimization of an ODE model.
    """

    @deprecate("Estimator", new="ODEHandler", version="2.19.1-zeta-fu1")
    def __init__(self, **kwargs):
        raise NotImplementedError("Please use ODEHandler()")


class Optimizer(Estimator):
    """
    This is deprecated. Please use Estimator class.
    """
    @deprecate("covsirphy.Estimator()", new="covsirphy.Estimator", version="2.17.0-eta")
    def __init__(self, **kwargs):
        super().__init__(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import contextlib
import copy
from datetime import timedelta
import json
from pathlib import Path
import re
import sys
import warnings
import numpy as np
import pandas as pd
from covsirphy.util.error import deprecate, ScenarioNotFoundError, UnExecutedError
from covsirphy.util.error import NotRegisteredMainError, NotRegisteredExtraError
from covsirphy.util.error import NotInteractiveError, UnExpectedValueRangeError
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy.visualization.line_plot import line_plot
from covsirphy.visualization.compare_plot import compare_plot
from covsirphy._deprecated.jhu_data import JHUData
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated._sir import SIR
from covsirphy._deprecated._sird import SIRD
from covsirphy._deprecated._sirf import SIRF
from covsirphy._deprecated.reg_handler import RegressionHandler
from covsirphy._deprecated.automl_handler import AutoMLHandler
from covsirphy._deprecated.data_handler import DataHandler
from covsirphy._deprecated.phase_tracker import PhaseTracker


class Scenario(Term):
    """
    Scenario analysis.

    Args:
        jhu_data (covsirphy.JHUData or None): object of records
        population_data (covsirphy.PopulationData or None): PopulationData object
        country (str): country name (must not be None)
        province (str or None): province name
        tau (int or None): tau value
        auto_complement (bool): if True and necessary, the number of cases will be complemented

    Note:
        @jhu_data and @population_data must be registered with Scenario.register() if not specified here.
    """

    @deprecate(old="Scenario", new="ODEScenario", version="2.24.0-xi")
    def __init__(self, jhu_data=None, population_data=None,
                 country=None, province=None, tau=None, auto_complement=True):
        # Area name
        if country is None:
            raise ValueError("@country must be specified.")
        self._area = JHUData.area_name(country, province)
        # Initialize data handler
        self._data = DataHandler(country=str(country), province=str(province or self.NA))
        self._data.switch_complement(whether=auto_complement)
        # ODE model
        self._model = None
        # Tau value
        self._tau = Validator(tau, "tau").tau(default=None)
        # Register datasets
        self._tracker_dict = {}
        self.register(jhu_data=jhu_data, population_data=population_data, extras=None)
        # Initialize parameter tracker
        with contextlib.suppress(NotRegisteredMainError):
            self._init_trackers()
        # Interactive (True) / script (False) mode
        self._interactive = hasattr(sys, "ps1")
        # Prediction of parameter values in the future phases: {name: RegressorBase)}
        self._reghandler_dict = {}

    def __getitem__(self, key):
        """
        Return the phase series object for the scenario.

        Args:
            key (str): scenario name

        Raises:
            ScenarioNotFoundError: the scenario is not registered

        Returns:
            covsirphy.PhaseTracker
        """
        if key in self._tracker_dict:
            return self._tracker_dict[key]
        raise ScenarioNotFoundError(key)

    def __setitem__(self, key, value):
        """
        Register a phase tracker.

        Args:
            key (str): scenario name
            value (covsirphy.PhaseTracker): phase tracker
        """
        Validator(value, "value").instance(PhaseTracker)
        self._tracker_dict[key] = value

    @property
    def first_date(self):
        """
        str: the first date of the records
        """
        return self._data.first_date

    @first_date.setter
    def first_date(self, date):
        self._data.timepoints(first_date=date, last_date=self._data.last_date, today=self._data.today)

    @property
    def last_date(self):
        """
        str: the last date of the records
        """
        return self._data.last_date

    @last_date.setter
    def last_date(self, date):
        try:
            Validator(date, "last_date").date(
                value_range=(Validator(self._data.today, "today").date(), None))
        except UnExpectedValueRangeError:
            today = date
        else:
            today = self._data.today
        self.timepoints(first_date=self._data.first_date, last_date=date, today=today)

    @property
    def today(self):
        """
        str: reference date to determine whether a phase is a past phase or a future phase
        """
        return self._data.today

    @today.setter
    def today(self, date):
        self.timepoints(first_date=self._data.first_date, last_date=self._data.last_date, today=date)

    @property
    def interactive(self):
        """
        bool: interactive mode (display figures) or not

        Note:
            When running scripts, interactive mode cannot be selected.
        """
        return self._interactive

    @interactive.setter
    def interactive(self, is_interactive):
        if not hasattr(sys, "ps1") and is_interactive:
            raise NotInteractiveError
        self._interactive = hasattr(sys, "ps1") and bool(is_interactive)

    def register(self, jhu_data=None, population_data=None, extras=None):
        """
        Register datasets.

        Args:
            jhu_data (covsirphy.JHUData or None): object of records
            population_data (covsirphy.PopulationData or None): PopulationData object (deprecated)
            extras (list[covsirphy.CleaningBase] or None): extra datasets

        Raises:
            TypeError: non-data cleaning instance was included
            UnExpectedValueError: instance of un-expected data cleaning class was included as an extra dataset
        """
        if population_data is not None:
            warnings.warn(
                "Scenario.register(population_data) was deprecated because population values are included in JHUData.",
                DeprecationWarning,
                stacklevel=2
            )
        self._data.register(jhu_data=jhu_data, population_data=population_data, extras=extras)
        if self._data.main_satisfied and not self._tracker_dict:
            self.timepoints()

    def timepoints(self, first_date=None, last_date=None, today=None):
        """
        Set the range of data and reference date to determine past/future of phases.

        Args:
            first_date (str or None): the first date of the records or None (min date of main dataset)
            last_date (str or None): the first date of the records or None (max date of main dataset)
            today (str or None): reference date to determine whether a phase is a past phase or a future phase

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data

        Note:
            When @today is None, the reference date will be the same as @last_date (or max date).
        """
        self._data.timepoints(first_date=first_date, last_date=last_date, today=today)
        self._init_trackers()

    def line_plot(self, df, show_figure=True, filename=None, **kwargs):
        """
        Display or save a line plot of the dataframe.

        Args:
            show_figure (bool): whether show figure when interactive mode or not
            filename (str or None): filename of the figure or None (not save) when script mode

        Note:
            When interactive mode and @show_figure is True, display the figure.
            When script mode and filename is not None, save the figure.
            When using interactive shell, we can change the modes by Scenario.interactive = True/False.
        """
        if self._interactive and show_figure:
            return line_plot(df=df, filename=None, **kwargs)
        if not self._interactive and filename is not None:
            return line_plot(df=df, filename=filename, **kwargs)

    def complement(self, **kwargs):
        """
        Complement the number of recovered cases, if necessary.

        Args:
            kwargs: the other arguments of JHUData.subset_complement()

        Returns:
            covsirphy.Scenario: self
        """
        self._data.switch_complement(whether=True, **kwargs)
        return self

    def complement_reverse(self):
        """
        Restore the raw records. Reverse method of covsirphy.Scenario.complement().

        Returns:
            covsirphy.Scenario: self
        """
        self._data.switch_complement(whether=False)
        return self

    def show_complement(self, **kwargs):
        """
        Show the details of complement that was (or will be) performed for the records.

        Args:
            kwargs: keyword arguments of JHUDataComplementHandler() i.e. control factors of complement

        Returns:
            pandas.DataFrame: as the same as JHUData.show_complement()
        """
        return self._data.show_complement(**kwargs)

    def _convert_variables(self, abbr, candidates):
        """
        Convert abbreviated variable names to complete names.

        Args:
            abbr (list[str] or str or None): variable names or abbreviated names
            candidates (list[str]): all candidates

        Returns:
            list[str]: complete names of variables

        Note:
            Selectable values of @abbr are as follows.
            - None: return default list, ["Infected", "Recovered", "Fatal"] (changed in the future)
            - list[str]: return the selected variables
            - "all": the all available variables
            - str: abbr, like "CIRF" (Confirmed/Infected/Fatal/Recovered), "CFR", "RC"
        """
        if abbr is None:
            return [self.CI, self.F, self.R]
        if abbr == "all":
            return Validator(candidates, "candidates").sequence()
        abbr_dict = {"C": self.C, "I": self.CI, "F": self.F, "R": self.R, }
        variables = list(abbr) if isinstance(abbr, str) else abbr
        variables = [abbr_dict.get(v, v) for v in variables]
        return Validator(variables, "variables").sequence(candidates=candidates)

    def records(self, variables=None, **kwargs):
        """
        Return the records as a dataframe.

        Args:
            variables (list[str] or str or None): variable names or abbreviated names
            kwargs: the other keyword arguments of Scenario.line_plot()

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            NotRegisteredExtraError: some variables are not included in the main datasets
            and no extra datasets were registered

        Returns:
            pandas.DataFrame

                Index
                    reset index
                Columns
                    - Date (pd.Timestamp): Observation date
                    - Columns set by @variables (int)

        Note:
            - Records with Recovered > 0 will be selected.
            - If complement was performed by Scenario.complement() or Scenario(auto_complement=True),
            The kind of complement will be added to the title of the figure.

        Note:
            Selectable values of @variables are as follows.
            - None: return default list, ["Infected", "Recovered", "Fatal"] (changed in the future)
            - list[str]: return the selected variables
            - "all": the all available variables
            - str: abbr, like "CIRF" (Confirmed/Infected/Recovered/Fatal), "CFR", "RC"
        """
        # Get necessary data for the variables
        all_df = self._data.records_all().set_index(self.DATE)
        variables = self._convert_variables(variables, all_df.columns.tolist())
        df = all_df.loc[:, variables]
        # Figure
        if self._data.complemented:
            title = f"{self._area}: Cases over time\nwith {self._data.complemented}"
        else:
            title = f"{self._area}: Cases over time"
        self.line_plot(df=df, title=title, y_integer=True, **kwargs)
        return df.reset_index()

    def records_diff(self, variables=None, window=7, **kwargs):
        """
        Return the number of daily new cases (the first discrete difference of records).

        Args:
            variables (list[str] or str or None): variable names or abbreviated names (as the same as Scenario.records())
            window (int): window of moving average, >= 1
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
                Index
                    - Date (pd.Timestamp): Observation date
                Columns
                    - Confirmed (int): daily new cases of Confirmed, if calculated
                    - Infected (int):  daily new cases of Infected, if calculated
                    - Fatal (int):  daily new cases of Fatal, if calculated
                    - Recovered (int):  daily new cases of Recovered, if calculated
        """
        window = Validator(window, "window").int(value_range=(1, None))
        df = self.records(variables=variables, show_figure=False).set_index(self.DATE)
        df = df.diff().dropna()
        df = df.rolling(window=window).mean().dropna().astype(np.int64)
        if self._data.complemented:
            title = f"{self._area}: Daily new cases\nwith {self._data.complemented}"
        else:
            title = f"{self._area}: Daily new cases"
        self.line_plot(df=df, title=title, y_integer=True, **kwargs)
        return df

    def _init_trackers(self):
        """
        Initialize dictionary of trackers.
        """
        tracker = PhaseTracker(self._data.records_main(), self.today, self._area)
        self._tracker_dict = {self.MAIN: tracker}

    def _tracker(self, name, template="Main"):
        """
        Ensure that the phases series is registered.
        If not registered, copy the template phase series.

        Args:
            name (str): phase series name
            template (str): name of template phase series

        Returns:
            covsirphy.ParamTracker

        Note:
            If regressors was registered by Scenario.fit(), the regressor will be removed.
        """
        # Registered
        if name in self._tracker_dict:
            return copy.deepcopy(self._tracker_dict[name])
        # Create it, if un-registered
        if template not in self._tracker_dict:
            raise ScenarioNotFoundError(template)
        tracker = copy.deepcopy(self._tracker_dict[template])
        self._tracker_dict[name] = tracker
        # Remove regressor
        if name in self._reghandler_dict:
            self._reghandler_dict.pop(name)
        return tracker

    @deprecate(old="Scenario.add_phase()", new="Scenario.add()")
    def add_phase(self, **kwargs):
        return self.add(**kwargs)

    def add(self, name="Main", end_date=None, days=None, model=None, tau=None, **kwargs):
        """
        Add a new phase.
        The start date will be the next date of the last registered phase.

        Args:
            name (str): phase series name, 'Main' or user-defined name
            end_date (str): end date of the new phase
            days (int): the number of days to add
            model (covsirphy.ModelBase or None): ODE model or None (not specified here)
            tau (int or None): tau value [min] or None (not specified here)
            kwargs: keyword arguments of ODE model parameters, not including tau value.

        Raises:
            ValueError: @end_date if smaller than the last end date of registered phases
            KeyError: model was registered, but some parameter values were not specified

        Returns:
            covsirphy.Scenario: self

        Note:
            If @end_date and @days are None, the end date will be the last date of the records.

        Note:
            When registered, ODE model and tau value will not be updated by @model and @tau.

        Note:
            If both of @end_date and @days were specified, @end_date will be used.

        Note:
            When ODE model and tau value has been or were registered, parameter values will be also added.
            Default values are that of the last phase. Er can change them with kwargs.
        """
        days = Validator(days, "days").int(value_range=(1, None))
        today = Validator(self._data.today).date()
        tracker = self._tracker(name)
        # Update ODE model
        if model is not None:
            self._model = Validator(model, "model").subclass(ModelBase)
        # Set tau value if has been un-set
        self._tau = self._tau or Validator(tau, "tau").tau(default=None)
        # Calculate start date and get parameter set
        if tracker:
            summary_df = tracker.summary()
            start = summary_df[self.END].max() + timedelta(days=1)
            pre_param_dict = summary_df.iloc[-1].to_dict()
            pre_param_dict.update(kwargs)
        else:
            start = Validator(self._data.first_date).date()
            pre_param_dict = kwargs.copy()
        # Add a new phase
        if end_date is None and days is not None:
            end = start + timedelta(days=days - 1)
        else:
            end = Validator(end_date, "end_date").date(default=today, value_range=(start, None))
        tracker.define_phase(start, end)
        # Set parameter values
        if None not in {self._model, self._tau} and (set(pre_param_dict) & set(self._model.PARAMETERS)):
            param_dict = {k: float(v) for (k, v) in pre_param_dict.items() if k in self._model.PARAMETERS}
            param_dict = Validator(param_dict, "param_dict").dict(required_keys=self._model.PARAMETERS, errors="raise")
            param_df = pd.DataFrame(param_dict, index=pd.date_range(start, end))
            tracker.set_ode(self._model, param_df, self._tau)
        # Update tracker of self
        self[name] = tracker
        return self

    def clear(self, name="Main", include_past=False, template="Main"):
        """
        Clear phase information.

        Args:
            name (str): scenario name
            include_past (bool): if True, past phases will be removed as well as future phases
            template (str): name of template scenario

        Returns:
            covsirphy.Scenario: self

        Note:
            If un-registered scenario name was specified, new scenario will be created.
            Future phases will be always deleted.
        """
        tracker = self._tracker(name, template=template)
        if include_past:
            self[name] = PhaseTracker(self._data.records_main(), self.today, self._area)
            return self
        df = tracker.summary()
        future_phases = df.loc[df[self.TENSE] == self.FUTURE].index.tolist()
        if not future_phases:
            return self
        start, end = tracker.parse_range(phases=future_phases)
        self[name] = tracker.remove_phase(start, end)
        return self

    def _delete_series(self, name):
        """
        Delete a scenario or initialise main scenario.

        Args:
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self
        """
        if name == self.MAIN:
            tracker = self._tracker(name)
            self[self.MAIN] = tracker.remove_phase(self._data.first_date, self._data.last_date)
        else:
            self._tracker_dict.pop(name)
        return self

    def delete(self, phases=None, name="Main"):
        """
        Delete phases.

        Args:
            phase (list[str] or None): phase names, or ['last']
            name (str): name of phase series

        Returns:
            covsirphy.Scenario: self

        Note:
            If @phases is None, the phase series will be deleted.
            When @phase is '0th', disable 0th phase. 0th phase will not be deleted.
            If the last phase is included in @phases, the dates will be released from phases.
            If the last phase is not included, the dates will be assigned to the previous phase.
        """
        # Clear main series or delete sub phase series
        if phases is None:
            return self._delete_series(name)
        # Delete phases
        tracker = self._tracker(name)
        last_phase = tracker.summary().index[-1]
        start, end = tracker.parse_range(phases=phases)
        f = {True: tracker.deactivate, False: tracker.remove_phase}
        self[name] = f[last_phase in phases or "last" in phases](start, end)
        return self

    def delete_matched(self, pattern):
        """
        Delete scenarios which matches to the given pattern.

        Arg:
            pattern (raw string): pattern to match

        Returns:
            covsirphy.Scenario: self
        """
        p = re.compile(pattern)
        names = [name for name in self._tracker_dict.keys() if p.match(name)]
        for name in names:
            self.delete(name=name)
        return self

    def _reverse(self, phases, name="Main"):
        """
        The status (enabled/disabled) will be reversed.

        Args:
            phase (list[str] or None): phase names or None (all enabled phases)
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        tracker = self._tracker(name)
        start, end = tracker.parse_range(phases=phases)
        self[name] = tracker.deactivate(start, end)
        return self

    def disable(self, phases, name="Main"):
        """
        The phases will be disabled and removed from summary.

        Args:
            phase (list[str] or None): phase names or None (all enabled phases)
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        df = self._tracker(name).summary()
        enabled_phases = df.loc[df[self.TENSE] == self.PAST].index.tolist()
        rev_phases = [ph for ph in phases if ph in enabled_phases]
        return self._reverse(phases=rev_phases, name=name)

    def enable(self, phases, name="Main"):
        """
        The phases will be enabled and appear in summary.

        Args:
            phase (list[str] or None): phase names or None (all disabled phases)
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        df = self._tracker(name).summary()
        enabled_phases = df.loc[df[self.TENSE] == self.PAST].index.tolist()
        rev_phases = [ph for ph in phases if ph not in enabled_phases]
        return self._reverse(phases=rev_phases, name=name)

    def combine(self, phases, name="Main", **kwargs):
        """
        Combine the sequential phases as one phase.
        New phase name will be automatically determined.

        Args:
            phases (list[str]): list of phases
            name (str, optional): name of phase series
            kwargs: keyword arguments of parameters

        Note:
            kwargs will be ignore when model and tau is not registered.

        Raises:
            TypeError: @phases is not a list

        Returns:
            covsirphy.Scenario: self
        """
        Validator(phases, name="phases").sequence()
        tracker = self._tracker(name)
        start, end = tracker.parse_range(phases=phases)
        tracker.define_phase(start, end)
        if kwargs and self._model is not None and self._tau is not None:
            kwargs = Validator(kwargs, "kwargs").dict(required_keys=self._model.PARAMETERS, errors="raise")
            param_df = pd.DataFrame(index=pd.date_range(start, end))
            for (param, value) in kwargs.items():
                param_df[param] = value
            tracker.set_ode(self._model, param_df, self._tau)
        self[name] = tracker
        return self

    def separate(self, date, name="Main", **kwargs):
        """
        Create a new phase with the change point.
        New phase name will be automatically determined.

        Args:
            date (pandas.Timestamp or str): change point, i.e. start date of the new phase
            name (str): scenario name
            kwargs: will be ignored

        Raises:
            ValueError: the date is close to one of the registered change dates

        Returns:
            covsirphy.Scenario: self
        """
        date = Validator(date, "date").date()
        tracker = self._tracker(name)
        # Get start date and end date which include the separation date
        df = tracker.summary()
        start_date = df.loc[df[self.START] < date, self.START].max()
        end_date = df.loc[df[self.END] > date, self.START].min()
        # Check separation date
        prohibited_dates = [
            start_date, start_date + timedelta(days=1),
            end_date - timedelta(days=-1), end_date
        ]
        if date in prohibited_dates:
            raise ValueError(
                f"Cannot be separated on {date} because this date is too close to registered change dates.")
        # Define phases
        tracker.define_phase(start_date, date - timedelta(days=1))
        tracker.define_phase(date, end_date)
        self[name] = tracker
        return self

    def _summary(self, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown

        Returns:
            pandas.DataFrame:
                - if @name not None, as the same as PhaseTracker.summary()
                - if @name is None, index will be phase series name and phase name

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        name = self.MAIN if len(self._tracker_dict) == 1 else name
        if name is not None:
            return self._tracker_dict[name].summary()
        dataframes = []
        for (_name, tracker) in self._tracker_dict.items():
            df = tracker.summary().rename_axis(self.PHASE)
            df[self.SERIES] = _name
            dataframes.append(df.reset_index())
        df = pd.concat(dataframes, ignore_index=True, sort=False)
        return df.set_index([self.SERIES, self.PHASE])

    def summary(self, columns=None, name=None):
        """
        Summarize the series of phases and return a dataframe.

        Args:
            name (str): phase series name
                - name of alternative phase series registered by Scenario.add()
                - if None, all phase series will be shown
            columns (list[str] or None): columns to show

        Returns:
            pandas.DataFrame:
            - if @name not None, as the same as PhaseTracker().summary()
            - if @name is None, index will be phase series name and phase name

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.

        Note:
            If @columns is None, all columns will be shown.

        Note:
            "Start" and "End" are string at this time.
        """
        df = self._summary(name=name).dropna(how="all", axis=1).fillna(self.NA)
        if df.empty:
            return pd.DataFrame(index=[self.TENSE, self.START, self.END, self.N])
        df[self.START] = df[self.START].dt.strftime(self.DATE_FORMAT)
        df[self.END] = df[self.END].dt.strftime(self.DATE_FORMAT)
        if columns is None:
            return df
        Validator(columns, "columns").sequence(df.columns.tolist())
        return df.loc[:, columns]

    def trend(self, min_size=None, force=True, name="Main", show_figure=True, filename=None, **kwargs):
        """
        Perform S-R trend analysis and set phases.

        Args:
            min_size (int or None): minimum value of phase length [days] (over 2) or None (equal to max of 7 and delay period)
            force (bool): if True, change points will be over-written
            name (str): phase series name
            show_figure (bool): if True, show the result as a figure
            filename (str): filename of the figure, or None (display)
            kwargs: keyword arguments of covsirphy.TrendDetector(), .TrendDetector.sr() and .trend_plot()

        Returns:
            covsirphy.Scenario: self

        Note:
            If @min_size is None, this will be thw max value of 7 days and delay period calculated with .estimate_delay() method.
        """
        # Arguments
        force = kwargs.pop("set_phases", force)
        # Minimum size of phases
        if min_size is None:
            try:
                delay, _ = self.estimate_delay(**Validator(kwargs, "keyword arguments").kwargs(
                    functions=self.estimate_delay, default=None))
            except KeyError:
                # Extra datasets are not registered
                delay = 7
            min_size = max(7, delay)
        kwargs["min_size"] = Validator(min_size, "min_size").int(value_range=(2, None))
        # S-R trend analysis
        tracker = self._tracker(name)
        if not self._interactive and filename is None:
            show_figure = False
        filename = None if self._interactive else filename
        self[name] = tracker.trend(force=force, show_figure=show_figure, filename=filename, **kwargs)
        # Disable 0th phase, if necessary
        if "include_init_phase" in kwargs:
            warnings.warn(
                "@include_init_phase was deprecated. Please use Scenario.disable('0th').",
                DeprecationWarning, stacklevel=2)
            self.disable(phases=["0th"], name=name)
        return self

    def estimate(self, model, phases=None, name="Main", **kwargs):
        """
        Perform parameter estimation for each phases.

        Args:
            model (covsirphy.ModelBase): ODE model
            phases (list[str]): list of phase names, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of ODEHandler(), ODEHandler.estimate_tau() and .estimate_param()

        Note:
            If @name phase was not registered, new tracker will be created.

        Note:
            If @phases is None, all past phase will be used.
        """
        self._model = Validator(model, "model").subclass(ModelBase)
        tracker = self._tracker(name)
        if self.TAU in kwargs:
            raise ValueError(
                "@tau must be specified when scenario = Scenario(), and cannot be specified here.")
        if phases is None:
            self._tau = self[name].estimate(self._model, tau=self._tau, **kwargs)
            return
        df = tracker.summary()
        all_phases = df.loc[df[self.TENSE] == self.PAST].index.tolist()
        self.disable(phases=all_phases, name=name)
        handler_kwargs = {"n_jobs": 1}
        handler_kwargs.update(kwargs)
        for ph in phases:
            self._reverse(phases=[ph], name=name)
            self._tau = self[name].estimate(self._model, tau=self._tau, **handler_kwargs)
            self._reverse(phases=[ph], name=name)
        self.enable(phases=all_phases, name=name)

    @ deprecate("Scenario.phase_estimator()", version="2.19.1-delta-fu1")
    def phase_estimator(self, **kwargs):
        """
        Deprecated. Return the estimator of the phase.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    @ deprecate("Scenario.estimate_history()", version="2.19.1-delta-fu1")
    def estimate_history(self, **kwargs):
        """
        Deprecated. Show the history of optimization.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def estimate_accuracy(self, phase, name="Main", **kwargs):
        """
        Show the accuracy as a figure.

        Args:
            phase (str): phase name, like 1st, 2nd...
            name (str): phase series name
            kwargs: keyword arguments of covsirphy.compare_plot()

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        variables = [self.CI, self.F, self.R]
        records_df = self.records(variables=variables, show_figure=False)
        tracker = self._tracker(name=name)
        sim_df = tracker.simulate()
        start, end = tracker.parse_range(phases=[phase])
        df = records_df.merge(sim_df, on=self.DATE, suffixes=("_actual", "_simulated"))
        df = df.set_index(self.DATE).loc[start:end]
        compare_plot(df, variables=variables, groups=["actual", "simulated"], **kwargs)

    def simulate(self, variables=None, name="Main", **kwargs):
        """
        Simulate ODE models with set/estimated parameter values and show it as a figure.

        Args:
            variables (list[str] or str or None): variable names or abbreviated names (as the same as Scenario.records())
            name (str): phase series name. If 'Main', main PhaseSeries will be used
            kwargs: the other keyword arguments of Scenario.line_plot() and PhaseTracker.parse_range()

        Returns:
            pandas.DataFrame
                Index
                    reset index
                Columns
                    - Date (pandas.Timestamp): Observation date
                    - Country (str): country/region name
                    - Province (str): province/prefecture/state name
                    - Variables of the main dataset (int): Confirmed etc.
        """
        tracker = self._tracker(name=name)
        # Variables to show
        variables = self._convert_variables(variables, candidates=self.VALUE_COLUMNS)
        # Specify range of dates
        start, end = tracker.parse_range(
            **Validator(kwargs, "keyword arguments").kwargs(functions=PhaseTracker.parse_range, default=None))
        sim_df = tracker.simulate().set_index(self.DATE).loc[start:end, variables]
        # Show figure
        start_dates = tracker.summary()[self.START].tolist()
        title = f"{self._area}: Simulated number of cases ({name} scenario)"
        self.line_plot(df=sim_df, title=title, y_integer=True, v=start_dates[1:], **kwargs)
        return sim_df.reset_index()

    def get(self, param, phase="last", name="Main"):
        """
        Get the parameter value of the phase.

        Args:
            param (str): parameter name (columns in self.summary())
            phase (str): phase name or 'last'
                - if 'last', the value of the last phase will be returned
            name (str): phase series name

        Returns:
            str or int or float

        Note:
            If 'Main' was used as @name, main PhaseSeries will be used.
        """
        df = self.summary(name=name)
        if param not in df.columns:
            raise KeyError(f"@param must be in {', '.join(df.columns)}.")
        if phase == "last":
            phase = df.index[-1]
        return df.loc[phase, param]

    @ deprecate(
        old="Scenario.param_history(targets: list)",
        new="Scenario.history(target: str)",
        version="2.7.3-alpha")
    def param_history(self, **kwargs):
        """
        Deprecated. Return subset of summary and show a figure to show the history.

        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def adjust_end(self):
        """
        Adjust the last end dates of the registered scenarios, if necessary.

        Returns:
            covsirphy.Scenario: self
        """
        # The current last end dates
        current_dict = {
            name: tracker.summary()[self.END].max()
            for (name, tracker) in self._tracker_dict.items()}
        # Adjusted end date
        adjusted = max(current_dict.values())
        for (name, _) in self._tracker_dict.items():
            with contextlib.suppress(UnExpectedValueRangeError):
                self.add(end_date=adjusted, name=name)
        return self

    def _describe(self):
        """
        Describe representative values.

        Returns:
            pandas.DataFrame
                Index
                    (int): scenario name
                Columns
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
        """
        _dict = {}
        for (name, tracker) in self._tracker_dict.items():
            # Predict the number of cases
            df = tracker.simulate().set_index(self.DATE)
            last_date = df.index[-1]
            # Max value of Infected
            max_ci = df[self.CI].max()
            argmax_ci = df[self.CI].idxmax().strftime(self.DATE_FORMAT)
            # Confirmed on the next date of the last phase
            last_c = df.loc[last_date, self.C]
            # Infected on the next date of the last phase
            last_ci = df.loc[last_date, self.CI]
            # Fatal on the next date of the last phase
            last_f = df.loc[last_date, self.F] if self.F in df.columns else None
            # Save representative values
            last_date_str = last_date.strftime(self.DATE_FORMAT)
            _dict[name] = {
                f"max({self.CI})": max_ci,
                f"argmax({self.CI})": argmax_ci,
                f"{self.C} on {last_date_str}": last_c,
                f"{self.CI} on {last_date_str}": last_ci,
                f"{self.F} on {last_date_str}": last_f,
            }
        return pd.DataFrame.from_dict(_dict, orient="index")

    def describe(self, with_rt=True, **kwargs):
        """
        Describe representative values.

        Args:
            with_rt (bool): whether show the history of Rt values
            kwargs: the other arguments will be ignored

        Returns:
            pandas.DataFrame:
                Index
                    str: scenario name
                Columns
                    - max(Infected): max value of Infected
                    - argmax(Infected): the date when Infected shows max value
                    - Confirmed({date}): Confirmed on the next date of the last phase
                    - Infected({date}): Infected on the next date of the last phase
                    - Fatal({date}): Fatal on the next date of the last phase
                    - nth_Rt etc.: Rt value if the values are not the same values
        """
        df = self._describe()
        if not with_rt or len(self._tracker_dict) == 1:
            return df
        # History of reproduction number
        rt_df = self.summary().reset_index().replace(self.NA, np.nan)
        rt_df = rt_df.pivot_table(index=self.SERIES, columns=self.PHASE, values=self.RT, aggfunc="last")
        rt_df = rt_df.fillna(self.NA)
        rt_df = rt_df.loc[:, rt_df.nunique() > 1]
        cols = sorted(rt_df, key=self.str2num)
        return df.join(rt_df[cols].add_suffix(f"_{self.RT}"), how="left")

    def track(self, with_actual=True, ref_name="Main", **kwargs):
        """
        Show values of parameters and variables in one dataframe.

        Args:
            with_actual (bool): if True, show actual number of cases will included as "Actual" scenario
            ref_name (str): name of reference scenario to specify phases and dates
            kwargs: keyword arguments of PhaseTracker.parse_range()

        Returns:
            pandas.DataFrame: tracking records
                Index
                    reset index
                Columns
                    - Scenario (str)
                    - Date (pandas.Timestamp)
                    - Confirmed (int): the number of confirmed cases
                    - Infected (int): the number of currently infected cases
                    - Fatal (int): the number of fatal cases
                    - Recovered (int): the number of recovered cases
                    - Susceptible (int): the number of susceptible cases
                    - Population (int)
                    - If available,
                        - Rt (float)
                        - parameter values (float)
                        - day parameter values (int)
        """
        unused_cols = [self.ODE, self.TRIALS, self.RUNTIME, self.TAU, *Evaluator.metrics()]
        # Tracking for scenarios
        dataframes = []
        append = dataframes.append
        for (name, tracker) in self._tracker_dict.items():
            df = tracker.track().drop(unused_cols, axis=1, errors="ignore")
            df.insert(0, self.SERIES, name)
            append(df)
        # Add actual records, if necessary
        if with_actual:
            df = self._data.records(extras=False)
            if self.SERIES not in df:
                df.insert(0, self.SERIES, self.ACTUAL)
            append(df)
        # Concat dataframes
        track_df = pd.concat(dataframes, axis=0, ignore_index=True, sort=False)
        track_df.insert(7, self.N, None)
        track_df[self.N] = track_df[[self.S, self.C]].sum(axis=1)
        # Specify date range
        ref_tracker = self._tracker(ref_name)
        start, end = ref_tracker.parse_range(
            **Validator(kwargs, "keyword arguments").kwargs(functions=PhaseTracker.parse_range, default=None))
        return track_df.loc[(track_df[self.DATE] >= start) & (track_df[self.DATE] <= end)]

    def _history(self, target, with_actual=True):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            with_actual (bool): if True and @target is a variable name, show actual number of cases

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp)
                Columns
                    {scenario name} (int or float): values of the registered scenario
        """
        # Include actual data or not
        with_actual = with_actual and target in self.VALUE_COLUMNS
        # Get tracking data
        df = self.track(with_actual=with_actual)
        if target not in df.columns:
            col_str = ", ".join(list(df.columns))
            raise KeyError(f"@target must be selected from {col_str}, but {target} was applied.")
        # Select the records of target variable
        return df.pivot_table(values=target, index=self.DATE, columns=self.SERIES, aggfunc="last")

    def history(self, target, with_actual=True, ref_name="Main", **kwargs):
        """
        Show the history of variables and parameter values to compare scenarios.

        Args:
            target (str): parameter or variable name to show (Rt, Infected etc.)
            with_actual (bool): if True and @target is a variable name, show actual number of cases
            ref_name (str): name of reference scenario to specify phases and dates
            kwargs: the other keyword arguments of Scenario.line_plot() and PhaseTracker.parse_range()

        Returns:
            pandas.DataFrame:
                Index
                    Date (pandas.Timestamp)
                Columns
                    {scenario name} (int or float): values of the registered scenario
        """
        df = self._history(target=target, with_actual=with_actual)
        df.dropna(subset=[col for col in df.columns if col != self.ACTUAL], inplace=True)
        ref_tracker = self._tracker(ref_name)
        start, end = ref_tracker.parse_range(
            **Validator(kwargs, "keyword arguments").kwargs(functions=PhaseTracker.parse_range, default=None))
        df = df.loc[start:end]
        # Y-label
        if target == self.RT:
            ylabel = self.RT_FULL
        elif target in self.VALUE_COLUMNS:
            ylabel = f"The number of {target.lower()} cases"
        else:
            ylabel = target
        # Title
        title = f"{self._area}: {ylabel} over time"
        # Dot lines
        start_dates = ref_tracker.summary()[self.START].tolist()
        # Show line plot
        self.line_plot(
            df=df, title=title, ylabel=ylabel, v=start_dates[1:], math_scale=False,
            h=1.0 if target == self.RT else None, **kwargs)
        return df

    def history_rate(self, params=None, name="Main", **kwargs):
        """
        Show change rates of parameter values in one figure.
        We can find the parameters which increased/decreased significantly.

        Args:
            params (list[str] or None): parameters to show
            name (str): phase series name
            kwargs: the other keyword arguments of Scenario.line_plot()

        Returns:
            pandas.DataFrame
        """
        df = self._tracker(name=name).track().set_index(self.DATE)
        cols = list(set(df.columns) & set(self._model.PARAMETERS))
        if params is not None:
            if not isinstance(params, (list, set)):
                raise TypeError(f"@params must be a list of parameters, but {params} were applied.")
            cols = list(set(cols) & set(params)) or cols
        df = df.loc[:, cols] / df.loc[df.index[0], cols]
        # Show figure
        f_date = df.index[0].strftime(self.DATE_FORMAT)
        title = f"{self._area}: {self._model.NAME} parameter change rates over time (1.0 on {f_date})"
        ylabel = f"Value per that on {f_date}"
        title = f"{self._area}: {ylabel} over time"
        tracker = self._tracker(self.MAIN)
        start_dates = tracker.summary()[self.START].tolist()
        self.line_plot(
            df=df, title=title, ylabel=ylabel, v=start_dates[1:], math_scale=False, **kwargs)
        return df

    def retrospective(self, beginning_date, model, control="Main", target="Target", **kwargs):
        """
        Perform retrospective analysis.
        Compare the actual series of phases (control) and
        series of phases with specified parameters (target).

        Args:
            beginning_date (str): when the parameter values start to be changed from actual values
            model (covsirphy.ModelBase): ODE model
            control (str): scenario name of control
            target (str): scenario name of target
            kwargs: keyword arguments of ODEHandler(), ODEHandler.estimate_tau() and .estimate_param()

        Note:
            When parameter values are not specified,
            actual values of the last date before the beginning date will be used.
        """
        param_dict = {k: v for (k, v) in kwargs.items() if k in model.PARAMETERS}
        # Control
        self.clear(name=control, include_past=True)
        self.trend(name=control, show_figure=False)
        with contextlib.suppress(ValueError):
            self.separate(date=beginning_date, name=control)
        self.estimate(model, name=control, **kwargs)
        # Target
        self.clear(name=target, include_past=False, template=control)
        df = self._tracker(target).summary()
        phases_changed = df.loc[df[self.START] >= pd.to_datetime(beginning_date)].index.tolist()
        self.delete(phases=phases_changed, name=target)
        self.add(name=target, **param_dict)
        self.estimate(model, phases=[phases_changed[0]], name=target, **kwargs)

    def score(self, variables=None, name="Main", **kwargs):
        """
        Evaluate accuracy of phase setting and parameter estimation.

        Args:
            variables (list[str] or None): variables to use in calculation
            name(str): phase series name. If 'Main', main PhaseSeries will be used
            kwargs: keyword arguments of covsirphy.Evaluator.score() and PhaseTracker.parse_range()

        Returns:
            float: score with the specified metrics (covsirphy.Evaluator.score())

        Note:
            If @variables is None, ["Infected", "Fatal", "Recovered"] will be used.

        Note:
            "Susceptible", "Confirmed", "Infected", "Fatal" and "Recovered" can be used in @variables.
        """
        tracker = self._tracker(name)
        # Variables to evaluate
        variables = Validator(variables, "variables").sequence(
            candidates=[self.S, self.C, self.CI, self.F, self.R], default=[self.CI, self.F, self.R])
        # Specify date range
        start, end = tracker.parse_range(
            **Validator(kwargs, "keyword arguments").kwargs(functions=PhaseTracker.parse_range, default=None))
        sim_df = tracker.simulate().set_index(self.DATE)
        rec_df = self.records(variables=variables, show_figure=False).set_index(self.DATE)
        # Evaluate
        evaluator = Evaluator(rec_df.loc[start:end], sim_df.loc[start:end, variables], how="inner")
        return evaluator.score(**Validator(kwargs, "keyword arguments").kwargs(functions=Evaluator.score, default=None))

    def estimate_delay(self, oxcgrt_data=None, indicator="Stringency_index",
                       target="Confirmed", percentile=25, limits=(7, 30), **kwargs):
        """
        Estimate delay period [days], assuming the indicator impact on the target value with delay.
        The average of representative value (percentile) and @min_size will be returned.

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset
            indicator (str): indicator name, a column of any registered datasets
            target (str): target name, a column of any registered datasets
            percentile (int): percentile to calculate the representative value, in (0, 100)
            limits (tuple(int, int)): minimum/maximum size of the delay period [days]
            kwargs: keyword arguments of DataHandler.estimate_delay()

        Raises:
            NotRegisteredMainError: JHUData was not registered
            SubsetNotFoundError: failed in subsetting because of lack of data
            UserWarning: failed in calculating and returned the default value (recovery period)

        Returns:
            tuple(int, pandas.DataFrame):
                - int: the estimated number of days of delay [day] (mode value)
                - pandas.DataFrame:
                    Index
                        reset index
                    Columns
                        - (int or float): column defined by @indicator
                        - (int or float): column defined by @target
                        - (int): column defined by @delay_name [days]

        Note:
            - Average recovered period of JHU dataset will be used as returned value when the estimated value was not in value_range.
            - @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).
        """
        min_size, max_days = limits
        # Register OxCGRT data
        if oxcgrt_data is not None:
            warnings.warn(
                "Please use Scenario.register(extras=[oxcgrt_data]) rather than Scenario.fit(oxcgrt_data).",
                DeprecationWarning, stacklevel=1)
            self.register(extras=[oxcgrt_data])
        # Un-used arguments
        if "value_range" in kwargs:
            warnings.warn("@value_range argument was deprecated.", DeprecationWarning, stacklevel=1)
        # Calculate delay values
        df = self._data.estimate_delay(
            indicator=indicator, target=target, min_size=min_size, delay_name="Period Length",
            **Validator(kwargs, "keyword arguments").kwargs(functions=DataHandler.estimate_delay, default=None))
        # Remove NAs and sort
        df.dropna(subset=["Period Length"], inplace=True)
        df.sort_values("Period Length", inplace=True)
        df.reset_index(inplace=True, drop=True)
        # Apply upper limit for delay period if max_days is set
        if max_days is not None:
            df = df[df["Period Length"] <= max_days]
        # Calculate representative value
        if df.empty:
            return (self._data.recovery_period(), df)
        # Calculate percentile
        Q1 = np.percentile(df["Period Length"], percentile, interpolation="midpoint")
        low_lim = min_size
        delay_period = int((low_lim + Q1) / 2)
        return delay_period, df

    @deprecate("Scenario.fit()", new="Scenario.predict()", version="2.23.0-alpha", ref="https://github.com/lisphilar/covid19-sir/issues/1006")
    def fit(self, oxcgrt_data=None, name="Main", delay=(7, 31), removed_cols=None, metric="MAPE", **kwargs):
        """
        Deprecated and merged to .predict().
        Fit regressors to predict the parameter values in the future phases,
        assuming that indicators will impact on ODE parameter values/the number of cases with delay.
        Please refer to covsirphy.RegressionHandler class.

        Args:
            oxcgrt_data (covsirphy.OxCGRTData): OxCGRT dataset, deprecated
            name (str): scenario name
            test_size (float): proportion of the test dataset of Elastic Net regression
            seed (int): random seed when splitting the dataset to train/test data
            delay (int or tuple(int, int) or None):
                - (int): delay period [days],
                - tuple(int, int): select the best value with grid search in this range, or
                - None: Scenario.estimate_delay() calculate automatically
            removed_cols (list[str] or None): list of variables to remove from X dataset or None (indicators used to estimate delay period)
            metric (str): metric name
            kwargs: keyword arguments of
                - sklearn.model_selection.train_test_split(),
                - RegressionHandler.feature_engineering(),
                - RegressorHandler.fit() and
                - Scenario.fit_accuracy()

        Raises:
            covsirphy.UnExecutedError: Scenario.estimate() or Scenario.add() were not performed

        Returns:
            dict: this is the same as covsirphy.RegressionHandler.to_dict()

        Note:
            @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).

        Note:
            Please refer to covsirphy.Evaluator.score() for metric names.

        Note:
            If @seed is included in kwargs, this will be converted to @random_state.

        Note:
            default values regarding sklearn.model_selection.train_test_split() are
            test_size=0.2, random_state=0, shuffle=False.
        """
        metric = kwargs.pop("metrics") if "metrics" in kwargs else metric
        # Clear the future phases
        self.clear(name=name, include_past=False)
        # Register OxCGRT data
        if oxcgrt_data is not None:
            warnings.warn(
                "Please use Scenario.register(extras=[oxcgrt_data]) rather than Scenario.fit(oxcgrt_data).",
                DeprecationWarning, stacklevel=1)
            self.register(extras=[oxcgrt_data])
        # ODE model
        if self._model is None:
            raise UnExecutedError(
                "Scenario.estimate() or Scenario.add()",
                details=f", specifying @model (covsirphy.SIRF etc.) and @name='{name}'.")
        # Create training/test dataset
        tracker = self._tracker(name=name)
        param_df = tracker.track().set_index(self.DATE)[self._model.PARAMETERS].dropna()
        try:
            records_df = self._data.records(main=True, extras=True).set_index(self.DATE)
        except NotRegisteredExtraError:
            raise NotRegisteredExtraError(
                "Scenario.register(jhu_data, population_data, extras=[...]) with extra datasets") from None
        records_df = records_df.loc[:, ~records_df.columns.isin(removed_cols or [])]
        data = param_df.join(records_df)
        # Set delay effect
        if delay is None:
            # Estimate delay period with Scenario.estimate_delay()
            delay, delay_df = self.estimate_delay(oxcgrt_data)
            records_df = records_df.loc[:, ~records_df.columns.isin(delay_df.columns)]
            data = param_df.join(records_df)
        elif not isinstance(delay, tuple):
            # Use specified delay value
            delay = Validator(delay, name="delay").int(value_range=(1, None))
        # Fit regression models
        handler = RegressionHandler(data=data, model=self._model, **kwargs)
        handler.feature_engineering(
            delay=delay, **Validator(kwargs, "keyword arguments").kwargs(functions=RegressionHandler.feature_engineering, default=None))
        handler.fit(
            metric=metric, **Validator(kwargs, "keyword arguments").kwargs(functions=RegressionHandler.fit, default=None))
        self._reghandler_dict[name] = handler
        # Accuracy
        self.fit_accuracy(name=name, metric=metric, **kwargs)
        # Return information
        return handler.to_dict(metric=metric)

    def _get_fit_handler(self, name="Main"):
        """
        Get the instance of RegressionHandler.

        Args:
            name (str): scenario name

        Raises:
            covsirphy.UnExecutedError: Scenario.fit() was not performed
        """
        if name not in self._reghandler_dict:
            raise UnExecutedError(f"Scenario.fit(name={name})")
        return self._reghandler_dict[name]

    def fit_accuracy(self, name="Main", metric="MAPE", show_figure=True, **kwargs):
        """
        Create a plot to show the accuracy of forecasting (predicted vs. actual parameter values).

        Args:
            name (str): scenario name
            metric (str): metric name, refer to covsirphy.Evaluator.score()
            show_figure (bool): whether show figure or not
            kwargs: keyword arguments of RegressionHandler.pred_true_plot()

        Returns:
            dict: this is the same as covsirphy.RegressionHandler.to_dict()
        """
        metric = kwargs.pop("metrics") if "metrics" in kwargs else metric
        handler = self._get_fit_handler(name=name)
        # Show predicted vs. actual plot
        if show_figure:
            handler.pred_actual_plot(
                metric=metric, **Validator(kwargs, "keyword arguments").kwargs(functions=RegressionHandler.pred_actual_plot, default=None))
        # Return information
        return handler.to_dict(metric=metric)

    def predict(self, days=30, method="multivariate_regression", name="Main", **kwargs):
        """
        Predict ODE parameter values with machine learning.

        Args:
            days (int): days to predict
            method (str): machine learning method name, refer to covsirphy.automl_handler.AutoMLHandler.predict().
            name (str): scenario name

        Returns:
            covsirphy.Scenario: self
        """
        if days is None or isinstance(days, list):
            return self.fit_predict(days=None, name="Main", **kwargs)
        # ODE model
        if self._model is None:
            raise UnExecutedError(
                "Scenario.estimate() or Scenario.add()",
                details=f"Note that we need to specify @model (covsirphy.SIRF etc.) and @name='{name}'.")
        # Create X/Y
        try:
            X = self._data.records(main=True, extras=True).set_index(self.DATE)
        except NotRegisteredExtraError:
            X = self._data.records(main=True, extras=False).set_index(self.DATE)
        tracker = self._tracker(name=name)
        Y = tracker.track().set_index(self.DATE)
        Y = Y.loc[Y.index <= self._data.today, self._model.PARAMETERS].dropna()
        # Prediction
        handler = AutoMLHandler(X, Y, model=self._model, days=days, **kwargs)
        handler.predict(method=method)
        phase_df = handler.summary()
        phase_df = phase_df.rename(
            columns={self.SERIES: "name", self.END: "end_date"}).drop([self.START, self.RT], axis=1)
        phase_df = phase_df.sort_values(["name", "end_date"], ignore_index=True)
        # Set new future phases
        for phase_dict in phase_df.to_dict(orient="records"):
            new_scenario = phase_dict["name"]
            if new_scenario not in self._tracker_dict:
                self.clear(name=new_scenario, template=name)
            self.add(**phase_dict)
        return self

    @deprecate("Scenario.fit_predict()", new="Scenario.predict()", version="2.23.0-alpha", ref="https://github.com/lisphilar/covid19-sir/issues/1006")
    def fit_predict(self, oxcgrt_data=None, days=None, name="Main", **kwargs):
        """
        Deprecated and renamed to .predict().
        Predict parameter values of the future phases using Elastic Net regression with OxCGRT scores,
        assuming that OxCGRT scores will impact on ODE parameter values with delay.
        New future phases will be added (over-written).

        Args:
            oxcgrt_data (covsirphy.OxCGRTData or None): OxCGRT dataset
            days (list[int]): list of days to predict or None (only the max value)
            name (str): scenario name
            kwargs: the other arguments of Scenario.fit() and Scenario.predict()

        Raises:
            covsirphy.UnExecutedError: Scenario.estimate() or Scenario.add() were not performed
            NotRegisteredExtraError: no extra datasets were registered

        Returns:
            covsirphy.Scenario: self

        Note:
            @oxcgrt_data argument was deprecated. Please use Scenario.register(extras=[oxcgrt_data]).
        """
        self.fit(
            oxcgrt_data=oxcgrt_data, name=name, **Validator(kwargs, "keyword arguments").kwargs(functions=Scenario.fit, default=None))
        handler = self._get_fit_handler(name=name)
        # Prediction with regression model
        df = handler.predict()
        # -> index=end_date (pandas.Timestamp), columns=parameter values
        df.index.name = "end_date"
        # Days to predict
        days = days or [len(df)]
        Validator(days, "days").sequence(candidates=list(range(1, len(df) + 1)))
        phase_df = df.reset_index().loc[[day - 1 for day in days], :]
        # Set new future phases
        for phase_dict in phase_df.to_dict(orient="records"):
            self.add(name=name, **phase_dict)
        return self

    def backup(self, filename):
        """
        Backup scenario information to a JSON file (so that we can use it with Scenario.restore()).

        Args:
            filename (str or pathlib.Path): JSON filename to backup the information
        """
        # Get information
        info_dict = {
            "area": self._area,
            "timepoint": {
                "first_date": self._data.first_date,
                "today": self._data.today,
                "last_date": self._data.last_date
            },
            "scenario": {}
        }
        for (name, tracker) in self._tracker_dict.items():
            summary_df = tracker.summary()
            for col in [self.START, self.END]:
                summary_df[col] = summary_df[col].dt.strftime(self.DATE_FORMAT)
            info_dict["scenario"][name] = summary_df.to_dict(orient="index")
        # Save information
        with Path(filename).open("w") as fh:
            json.dump(info_dict, fh, indent=4)

    def restore(self, filename):
        """
        Restore scenario information with a JSON file (written by Scenario.backup()).

        Args:
            filename (str or pathlib.Path): JSON file to read

        Returns:
            covsirphy.Scenario: self

        Note:
            If keyword arguments of Scenario.timepoints() are available as the values of "timepoint" key,
            timepoints will be restored.
        """
        # Read the file
        with Path(filename).open("r") as fh:
            info_dict = json.load(fh)
        # Check area name
        if "area" not in info_dict or info_dict["area"] != self._area:
            raise ValueError(f"Area name (value of 'area' key) must be {self._area},\n{filename}")
        # Set timepoints
        self.timepoints(**Validator(info_dict.get("timepoint", {})).kwargs(functions=self.timepoints, default=None))
        # Set scenarios and phases
        snl_dict = info_dict.get("scenario", {})
        model_dict = {SIR.NAME: SIR, SIRD.NAME: SIRD, SIRF.NAME: SIRF}
        for (name, summary_dict) in snl_dict.items():
            self.clear(name=name, include_past=True)
            for phase_dict in summary_dict.values():
                add_dict = phase_dict.copy()
                add_dict["end_date"] = add_dict.pop(self.END)
                if self.ODE in add_dict:
                    model_name = add_dict.pop(self.ODE)
                    add_dict["model"] = model_dict[model_name]
                self.add(name=name, **add_dict)
        return self

    def rename(self, old, new):
        """
        Rename the given scenario names with a new one.

        Args:
            old (str): old name
            new (str): new name

        Raises:
            ScenarioNotFoundError: the old scenario does not exist
            ValueError: the new scenario name has already exist

        Returns:
            covsirphy.Scenario: self
        """
        if old not in self._tracker_dict:
            raise ScenarioNotFoundError(name=old)
        if new in self._tracker_dict:
            raise ValueError(f"@new={new}, but this scenario has already exist.")
        self._tracker(name=new, template=old)
        self._delete_series(name=old)
        return self

    def represent(self, q, variable, date=None, included=None, excluded=None):
        """
        Return the names of representative scenarios using quantiles of the variable on on the date.

        Args:
            q (list[float] or float): quantiles
            variable (str): reference variable, Confirmed, Infected, Fatal or Recovered
            date (str): reference date or None (the last end date in the all scenarios)
            included (list[str] or None): included scenarios or None (all included)
            excluded (list[str] or None): excluded scenarios or None (no scenarios not excluded)

        Raises:
            ValueError: the end dates of the last phase is not aligned

        Returns:
            list[float] or float: the nearest scenarios which has the values at the given quantiles

        Note:
            Dimension of returned object corresponds to the type of @q.
        """
        quantiles = [q] if isinstance(q, float) else Validator(q, "q").sequence()
        Validator([variable], "variable").sequence(candidates=self.VALUE_COLUMNS)
        # Target scenarios to represent
        all_set = set(self._tracker_dict.keys())
        in_set = all_set if included is None else set(Validator(included, "included").sequence())
        ex_set = set() if excluded is None else set(Validator(excluded, "excluded").sequence())
        scenarios = list(all_set & (in_set) - ex_set)
        # Get simulation data of the variable of the target scenarios
        sim_dict = {name: self._tracker_dict[name].simulate().set_index(self.DATE)[variable] for name in scenarios}
        sim_df = pd.DataFrame(sim_dict)
        if sim_df.isna().to_numpy().sum():
            raise ValueError(
                "The end dates of the last phases must be aligned. Scenario.adjust_end() method may fix this issue.")
        # Find representative scenario
        date_obj = Validator(date, "date").date(default=sim_df.index[-1])
        series = sim_df.loc[sim_df.index == date_obj].squeeze()
        values = series.quantile(q=quantiles, interpolation="nearest")
        return [series[series == v].index.tolist()[0] for v in values]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from matplotlib import pyplot as plt
from matplotlib.ticker import ScalarFormatter
import numpy as np
from covsirphy.util.error import deprecate
from covsirphy.util.validator import Validator
from covsirphy.visualization.vbase import find_args
from covsirphy.visualization.line_plot import LinePlot


class TrendPlot(LinePlot):
    """
    Create line plot with actual values for S-R trend analysis.

    Args:
        filename (str or None): filename to save the figure or None (display)
        bbox_inches (str): bounding box in inches when creating the figure
        kwargs: the other arguments of matplotlib.pyplot.savefig()
    """

    @deprecate(old="TrendPlot", new="Dynamics.detect()", version="2.24.0-xi")
    def __init__(self, filename=None, bbox_inches="tight", **kwargs):
        self._filename = filename
        self._savefig_dict = {"bbox_inches": bbox_inches, **kwargs}
        # Properties
        self._title = ""
        self._variables = []
        _, self._ax = plt.subplots(1, 1)

    def __enter__(self):
        return super().__enter__()

    def __exit__(self, *exc_info):
        return super().__exit__(*exc_info)

    def plot(self, data, actual_col="Actual"):
        """
        Plot chronological change of the data with multiple lines.

        Args:
            data (pandas.DataFrame): data to show
                Index
                    x values
                Columns
                    - column defined by @actual_col, actual values for y-axis
                    - the other arguments will be assumed as predicted values for y-axis
            actual_col (str): column name for y-axis
        """
        Validator(data, "data").dataframe(columns=[actual_col])
        predicted_cols = [col for col in data.columns if col != actual_col]
        self._variables = data.columns.tolist()
        # Scatter plot (actual values)
        self._ax.plot(
            data.index, data[actual_col],
            label=actual_col, color="black", marker=".", markeredgewidth=0, linewidth=0)
        # Plot lines
        for col in predicted_cols:
            self._ax.plot(data.index, data[col], label=col)
        plt.tight_layout()

    def x_axis(self, xlabel=None, xlim=(None, None)):
        """
        Set x axis.

        Args:
            xlabel (str or None): x-label
            xlim (tuple(int or float, int or float)): limit of x domain

        Note:
            When xlim[0] is None and lower x-axis limit determined by matplotlib automatically is lower than 0,
            lower x-axis limit will be set to 0.
        """
        # Label
        self._ax.set_xlabel(xlabel)
        # limit
        self._ax.set_xlim(max(self._ax.get_xlim()[0], xlim[0] or 0), xlim[1])
        # Integer scale
        fmt = ScalarFormatter(useOffset=False)
        fmt.set_scientific(False)
        self._ax.xaxis.set_major_formatter(fmt)

    def y_axis(self, ylabel=None):
        """
        Set x axis.

        Args:
            ylabel (str or None): y-label
        """
        # Label
        self._ax.set_ylabel(ylabel)
        # Scale
        try:
            self._ax.set_yscale("log", base=10)
        except Exception:
            # Matplotlib version < 3.3
            self._ax.set_yscale("log", basey=10)
        # Log scale
        # Delete y-labels of log-scale (minor) axis
        plt.setp(self._ax.get_yticklabels(minor=True), visible=False)
        self._ax.tick_params(left=False, which="minor")
        # Set new y-labels of major axis
        ymin, ymax = self._ax.get_ylim()
        ydiff_scale = int(np.log10(ymax - ymin))
        yticks = np.linspace(
            round(ymin, - ydiff_scale),
            round(ymax, - ydiff_scale), 5, dtype=np.int64)
        self._ax.set_yticks(yticks)
        fmt = ScalarFormatter(useOffset=False)
        fmt.set_scientific(False)
        self._ax.yaxis.set_major_formatter(fmt)

    def legend(self, bbox_to_anchor=(0.5, -0.5), bbox_loc="lower center", ncol=7, **kwargs):
        """
        Set legend.

        Args:
            bbox_to_anchor (tuple(int or float, int or float)): distance of legend and plot
            bbox_loc (str): location of legend
            ncol (int or None): the number of columns that the legend has
            kwargs: keyword arguments of matplotlib.pyplot.legend()
        """
        super().legend(bbox_to_anchor=bbox_to_anchor, bbox_loc=bbox_loc, ncol=ncol, **kwargs)


@deprecate(old="trend_plot()", new="Dynamics.detect()", version="2.24.0-xi")
def trend_plot(df, title=None, filename=None, show_legend=True, **kwargs):
    """
    Wrapper function: show chronological change of the data.

    Args:
        df (pandas.DataFrame): data to show
            Index
                x values
            Columns
                - column defined by @actual_col, actual values for y-axis
                - columns defined by @predicted_cols, predicted values for y-axis
        actual_col (str): column name for y-axis
        predicted_cols (list[str]): list of columns which have predicted values
        title (str): title of the figure
        filename (str or None): filename to save the figure or None (display)
        show_legend (bool): whether show legend or not
        kwargs: keyword arguments of the following classes and methods.
            - covsirphy.TrendPlot() and its methods,
            - matplotlib.pyplot.savefig() and matplotlib.pyplot.legend()
    """
    with TrendPlot(filename=filename, **find_args(plt.savefig, **kwargs)) as tp:
        tp.title = title
        tp.plot(data=df, **find_args([TrendPlot.plot], **kwargs))
        # Axis
        tp.x_axis(**find_args([TrendPlot.x_axis], **kwargs))
        tp.y_axis(**find_args([TrendPlot.y_axis], **kwargs))
        # Vertical/horizontal lines
        tp.line(**find_args([TrendPlot.line], **kwargs))
        # Legend
        if show_legend:
            tp.legend(**find_args([TrendPlot.legend, plt.legend], **kwargs))
        else:
            tp.legend_hide()


@deprecate(old="covsirphy.line_plot_multiple()", new="covsirphy.trend_plot()")
def line_plot_multiple(df, x_col, actual_col, predicted_cols, **kwargs):
    """
    Show multiple line graph of chronological change with actual plots.
    This function was deprecated. Please use covsirphy.trend_plot() function.

    Args:
        df (pandas.DataFrame): data
            Index
                Date (pandas.TimeStamp): Observation date
            Columns
                - column defined by @x_col, values for x-axis
                - column defined by @actual_col, actual values for y-axis
                - columns defined by @predicted_cols, predicted values for y-axis
        x_col (str): column name for x-axis
        actual_col (str): column name for y-axis
        kwargs: any other arguments of covsirphy.trend_plot()
    """
    return trend_plot(df.set_index(x_col), actual_col=actual_col, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from covsirphy.__version__ import __version__

__citation__ = "Hirokazu Takaya and CovsirPhy Development Team (2020-2022), " \
    f"CovsirPhy version {__version__}: " \
    "Python library for COVID-19 analysis with phase-dependent SIR-derived ODE models, " \
    "https://github.com/lisphilar/covid19-sir"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import math
import optuna
from optuna.samplers import TPESampler
from covsirphy.util.error import NAFoundError
from covsirphy.util.evaluator import Evaluator
from covsirphy.util.stopwatch import StopWatch
from covsirphy.util.validator import Validator
from covsirphy.util.term import Term
from covsirphy._deprecated._mbase import ModelBase
from covsirphy._deprecated.ode_solver import _ODESolver


class _ParamEstimator(Term):
    """
    Estimate ODE parameter values with records.

    Args:
        model (covsirphy.ModelBase): ODE model
        data (pandas.DataFrame):
            Index
                reset index
            Columns
                - Date (pd.Timestamp): Observation date
                - Susceptible(int): the number of susceptible cases
                - Infected (int): the number of currently infected cases
                - Fatal(int): the number of fatal cases
                - Recovered (int): the number of recovered cases
        tau (int): tau value [min]
        metric (str): metric to minimize
        quantiles (tuple(int, int)): quantiles to cut parameter range, like confidence interval
    """
    PRUNER_DICT = {
        "hyperband": optuna.pruners.HyperbandPruner,
        "median": optuna.pruners.MedianPruner,
        "threshold": optuna.pruners.ThresholdPruner,
        "percentile": optuna.pruners.PercentilePruner,
    }

    def __init__(self, model, data, tau, metric, quantiles):
        self._model = Validator(model, "model").subclass(ModelBase)
        Validator(data, "data").dataframe(columns=self.DSIFR_COLUMNS)
        self._tau = Validator(tau, "tau").tau(default=None)
        if self._tau is None:
            raise NAFoundError("tau", None)
        self._metric = Validator([metric], "metric").sequence(candidates=Evaluator.metrics())[0]
        # time steps (index), variables of the model
        df = model.convert(data, tau)
        self._taufree_df = df.copy()
        # Initial values
        self._y0_dict = df.iloc[0].to_dict()
        # Total population
        self._population = df.iloc[0].sum()
        # Step numbers
        self._step_n = df.index.max()
        # Parameter range
        self._range_dict = model.guess(data, tau, q=quantiles)
        # Max values of the variables
        self._max_dict = {v: df[v].max() for v in model.VARIABLES}

    def run(self, check_dict, study_dict):
        """
        Perform parameter estimation of the ODE model, not including tau.

        Args:
            check_dict (dict[str, object]): setting of validation
                - timeout (int): timeout of optimization
                - timeout_iteration (int): timeout of one iteration
                - tail_n (int): the number of iterations to decide whether score did not change for the last iterations
                - allowance (tuple(float, float)): the allowance of the max predicted values
            study_dict (dict[str, object]): setting of optimization study
                - pruner (str): kind of pruner (hyperband, median, threshold or percentile)
                - upper (float): works for "threshold" pruner, intermediate score is larger than this value, it prunes
                - percentile (float): works for "Percentile" pruner, the best intermediate value is in the bottom percentile among trials, it prunes
                - constant_liar (bool): whether use constant liar to reduce search effort or not

        Returns:
            dict(str, object):
                - Rt (float): phase-dependent reproduction number
                - (dict(str, float)): estimated parameter values
                - (dict(str, int or float)): day parameters, including 1/beta [days]
                - {metric}: score with the estimated parameter values
                - Trials (int): the number of trials
                - Runtime (str): runtime of optimization

        Note:
            Please refer to covsirphy.Evaluator.score() for metric names.
        """
        timeout = check_dict["timeout"]
        timeout_iteration = check_dict["timeout_iteration"]
        tail_n = check_dict["tail_n"]
        allowance = check_dict["allowance"]
        # Initialize optimization
        study = self.init_study(**study_dict)
        # The number of iterations
        iteration_n = math.ceil(timeout / timeout_iteration)
        stopwatch = StopWatch()
        # Optimization
        scores = []
        param_dict = {}
        for _ in range(iteration_n):
            # Run iteration
            study.optimize(self.objective, n_jobs=1, timeout=timeout_iteration)
            param_dict = study.best_params.copy()
            # If score did not change in the last iterations, stop running
            scores.append(self._score(**param_dict))
            if len(scores) >= tail_n and len(set(scores[-tail_n:])) == 1:
                break
            # Check max values are in the allowance
            if self.is_in_allowance(allowance, **param_dict):
                break
        model_instance = self._model(self._population, **param_dict)
        return {
            self.RT: model_instance.calc_r0(),
            **param_dict.copy(),
            **model_instance.calc_days_dict(self._tau),
            self._metric: self._score(**param_dict),
            self.TRIALS: len(study.trials),
            self.RUNTIME: stopwatch.stop_show(),
        }

    def init_study(self, pruner, **kwargs):
        """
        Initialize Optuna study.

        Args:
            pruner (str): Hyperband, Median, Threshold or Percentile
            kwargs: keyword arguments of pruners and TPESampler

        Returns:
            optuna.study.Study
        """
        v = Validator(kwargs, "keyword arguments")
        pruner_class = self.PRUNER_DICT.get(pruner.lower(), optuna.pruners.ThresholdPruner)
        pruner = pruner_class(**v.kwargs(functions=pruner_class, default=None))
        sampler = TPESampler(**v.kwargs(functions=TPESampler, default=None))
        return optuna.create_study(direction="minimize", sampler=sampler, pruner=pruner)

    def objective(self, trial):
        """
        Objective function to minimize.
        Score will be calculated with the data and metric.

        Args:
            trial (optuna.trial): a trial of the study

        Returns:
            float: score
        """
        param_dict = {}
        for (k, v) in self._range_dict.items():
            try:
                param_dict[k] = trial.suggest_uniform(k, *v)
            except OverflowError:
                param_dict[k] = trial.suggest_uniform(k, 0, 1)
        return self._score(**param_dict)

    def _score(self, **kwargs):
        """
        Objective function to minimize.
        Score will be calculated the data and metric.

        Args:
            kwargs: values of non-dimensional model parameters, including rho and sigma

        Returns:
            float: score
        """
        # Simulate with applied parameter values
        solver = _ODESolver(model=self._model, **kwargs)
        sim_df = solver.run(step_n=self._step_n, **self._y0_dict)
        # The first variable (Susceptible) will be ignored in score calculation
        taufree_df = self._taufree_df.loc[:, self._taufree_df.columns[1:]]
        sim_df = sim_df.loc[:, sim_df.columns[1:]]
        # Calculate score
        evaluator = Evaluator(taufree_df, sim_df, how="inner", on=None)
        return evaluator.score(metric=self._metric)

    def is_in_allowance(self, allowance, **kwargs):
        """
        Return whether all max values of estimated values are in allowance or not.

        Args:
            allowance (tuple(float, float)): the allowance of the predicted value
            kwargs: values of non-dimensional model parameters, including rho and sigma

        Returns:
            (bool): True when all max values of predicted values are in allowance
        """
        # Get max values with estimated parameter values
        solver = _ODESolver(model=self._model, **kwargs)
        sim_df = solver.run(step_n=self._step_n, **self._y0_dict)
        sim_max_dict = {v: sim_df[v].max() for v in self._model.VARIABLES}
        # Check all max values are in allowance
        allowance0, allowance1 = allowance
        ok_list = [
            a * allowance0 <= p <= a * allowance1 for (a, p) in zip(self._max_dict.values(), sim_max_dict.values())]
        return all(ok_list)
